/* The controller for picking up packaged component instances on demand.
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.controller;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;

import com.bbn.morph.evaluator.MainEvaluationController;
import com.bbn.morph.executer.StressInputs;
import com.bbn.morph.model.ComponentInstance;
import com.bbn.morph.model.ComponentLibrary;
import com.bbn.morph.model.CompositeModel;
import com.bbn.morph.model.InvalidModelException;
import com.bbn.morph.model.Measurement;
import com.bbn.morph.model.ModelBody;
import com.bbn.morph.model.PhysicalModel;
import com.bbn.morph.model.RobotAttributeValues;
import com.bbn.morph.ui.InformationReporter;
import com.bbn.morph.view.AttributeValueView;
import com.bbn.morph.view.DesiredAttributeValueView;

public class DiscreteUpdateController {

  private PhysicalModel m_model;
  private CompositeModel m_compositeModel;
  private InformationReporter m_reporter = null;
  
  public DiscreteUpdateController(CompositeModel compositeModel) {
    m_model = compositeModel.getRobotModel();
    m_compositeModel = compositeModel;
  }
  /**
   * @param reporter the information reporter to use to report information
   */
  public void setInformationReporter(InformationReporter reporter) {
    m_reporter  = reporter;
  }
  
  /**
   * Iterate picking components until all components are chosen or the
   * maximum number of iterations is reached.
   * @param maxIterations the maximum number of iterations to take
   * @throws InvalidModelException 
   */
  public void chooseAllComponents(int maxIterations) throws InvalidModelException {
    for (int i=1; i<=maxIterations; i++) {
      boolean updated = runIteration();
      if (!updated)
        return;
    }
  }
  
  /**
   * Pick one component . 
   * @throws InvalidModelException 
   */
  public void chooseOneComponent() throws InvalidModelException {
    runIteration();
  }
  

  
  /**
   * Run an iteration of the discrete update
   * @return true iff a component was updated
   * @throws InvalidModelException 
   */
  private boolean runIteration() throws InvalidModelException {
    List<ModelBody> needToChooseComponents = new ArrayList<ModelBody>();
    for (ModelBody body : m_model.getModelBodies()) {
      if (body.hasComponentLibrary() && !body.isComponentFixed())
        needToChooseComponents.add(body);
    }
    if (needToChooseComponents.isEmpty())
      return false;
    
    List<ModelChange> possibleChanges = new ArrayList<ModelChange>();
    for (ModelBody baseBody : needToChooseComponents) {
      ComponentLibrary library = baseBody.getComponentLibrary();
      // get all the possible components given the base body
      for (ComponentInstance component : library.getComponents(baseBody)) {
        PhysicalModel copy = m_model.copy();
        ModelBody bodyInCopy = findBodyInCopy(baseBody, copy);
        possibleChanges.add(new ModelChange(bodyInCopy, component, copy));
      }
    }
    
    // TODO this is the part we should run off on the cluster and then do a join on
    for (ModelChange candidateChange : possibleChanges) {
      PhysicalModel model = candidateChange.getModifiedModel();
      // actually put the component changes into the model
      updateModel(model, candidateChange.getModelBody(), candidateChange.getComponentInstance());
      
      CompositeModel newCompositeModel = new CompositeModel(model, 
          m_compositeModel.getEnvironmentModel(), m_compositeModel.getFbModel(),
          m_compositeModel.getEvaluatorFilename());
      AttributeValueView view = new AttributeValueView(newCompositeModel.getRobotModel());
      DesiredAttributeValueView dView = new DesiredAttributeValueView(newCompositeModel.getDesiredRobotModel());
      MainEvaluationController evalControl = new MainEvaluationController(newCompositeModel, view, dView);
      candidateChange.setStresses(evalControl.getStressInputs());
    }
    
    // now we need to pick the candidate change with the lowest stress:
    ModelChange lowest = pickCandidate(possibleChanges);
    // actually change the actual model!
    updateModel(m_model, lowest.getModelBody(), lowest.getComponentInstance());
    
    if (m_reporter != null)
      m_reporter.reportInformation("Chose "+lowest.getComponentInstance().getName()+" for component "+lowest.getModelBody().getId());

    return true;
  }
  
  private ModelBody findBodyInCopy(ModelBody body, PhysicalModel copy) {
    return copy.getIDMap().get(body.getId());
  }
  
  /**
   * Pick the {@link ModelChange} with the lowest stress value, or if 
   * several are equal pick a random one from among the lowest.
   * @param possibleChanges the candidate changes
   * @return a candidate change with the lowest stress value
   */
  private ModelChange pickCandidate(List<ModelChange> possibleChanges) {
    List<ModelChange> leastStressed = new ArrayList<ModelChange>();
    Double lowestStress = Double.MAX_VALUE;
    
    for (ModelChange change : possibleChanges) {
      double totalStress = change.getTotalStress();
      if (totalStress <= lowestStress) {
        if (totalStress != lowestStress) {
          leastStressed.clear();
          lowestStress = totalStress;
        }
        leastStressed.add(change);
      }
    }
    
    // now check the size of least stressed -- if there is only one return it
    if (leastStressed.size() == 1)
      return leastStressed.get(0);
    else {
      // return a random element:
      Random rand = new Random();
      int i = rand.nextInt(leastStressed.size());
      return leastStressed.get(i);
    }
  }

  private void updateModel(PhysicalModel model, ModelBody modelBody,
      ComponentInstance componentInstance) {
    ModelBody bodyToUpdate = findBodyInCopy(modelBody, model);
    List<RobotAttributeValues> newAttributes = componentInstance.getAttributes();
    for (RobotAttributeValues attr: newAttributes) {
      Measurement delta = attr.getMeasurement().subtract(bodyToUpdate.getAttributeValue(attr.getAttribute()));
      model.updateRobotAttributeFromDeltas(new RobotAttributeValues(bodyToUpdate.getId()+"."+attr.getAttribute(), delta));
    }
    bodyToUpdate.setComponentInstance(componentInstance);
  }

  private class ModelChange {
    private ModelBody m_body;
    private ComponentInstance m_component;
    private PhysicalModel m_modifiedModel;
    private Set<StressInputs> m_stressInputs = new HashSet<StressInputs>();
    private double totalStress = 0;
    
    public ModelChange(ModelBody body, ComponentInstance component,
        PhysicalModel modifiedModel) {
      m_body = body;
      m_component = component;
      m_modifiedModel = modifiedModel;
    }
    
    public void setStresses(Set<StressInputs> stressInputs) {
      m_stressInputs.addAll(stressInputs);
      for (StressInputs stress : stressInputs) {
        for (Double s: stress.values())
          totalStress += (Math.abs(s));
      }
    }

    public ModelBody getModelBody() {
      return m_body;
    }
    
    public ComponentInstance getComponentInstance() {
      return m_component;
    }
    
    public PhysicalModel getModifiedModel() {
      return m_modifiedModel;
    }
    
    public double getTotalStress() {
      return totalStress;
    }
  }
  
}
