/* Parse evaluators from a stream
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */

package com.bbn.morph.evaluator;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.bbn.morph.model.CompositeModel;
import com.bbn.morph.model.InvalidModelException;
import com.bbn.morph.view.AttributeValueView;
import com.bbn.morph.view.DesiredAttributeValueView;

public class EvaluatorParser {

  /**
   * Parse the input stream (in XML format) into a list of {@link Evaluator}s.
   * Evaluators are parsed iff there is no active attribute or it is set to true.
   * @param stream the stream of XML containing the evaluators
   * @param cModel the composite model
   * @param model a view of the model 
   * @param dModel a view of the desired model
   * @return a list of evaluators
   * @throws InvalidModelException if the stream contains badly constructed evaluators or XML
   */
  public static List<Evaluator> parseEvaluators(InputStream stream, CompositeModel cModel,
      AttributeValueView model, DesiredAttributeValueView dModel) 
          throws InvalidModelException {
    DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();

    if (stream == null) {
      throw new InvalidModelException("Invalid (null) stream for Evaluators.");
    }
    
    try {
      // Using factory get an instance of document builder
      DocumentBuilder db = dbf.newDocumentBuilder();

      // Parse using builder to get DOM representation of the XML stream
      Document dom = db.parse(stream);

      Element rootElement = dom.getDocumentElement();
      if (!rootElement.getNodeName().equals("evaluators")) {
        throw new InvalidModelException("Expected root node called \"evaluators\", instead found: "+rootElement.getNodeName());
      }

      List<Evaluator> evaluators = new ArrayList<Evaluator>();
      NodeList childNodes = rootElement.getChildNodes();
      for (int i=0; i<childNodes.getLength(); i++) {
        Node node = childNodes.item(i);
        if (node.getNodeType() == Node.ELEMENT_NODE) {
          Element element = (Element) node;
          Evaluator e = createEvaluator(element, cModel, model, dModel);
          if (e != null)
            evaluators.add(e);
        }
      }

      return evaluators;
    } catch(Exception e) {
      throw new InvalidModelException(e.getMessage());
    }
  }

  private static Evaluator createEvaluator(Element element, CompositeModel cModel,
      AttributeValueView model, DesiredAttributeValueView dModel) 
          throws InvalidModelException {
    String nodeName = element.getNodeName();
    if (nodeName != "evaluator")
      throw new InvalidModelException("Expected element called \"evaluator\", instead found: "+nodeName);
    // active if there is no active attribute or the attribute is true
    boolean isActive = (!element.hasAttribute("active") || 
        Boolean.parseBoolean(element.getAttribute("active")));
    if (isActive) {
      String evaluatorType = element.getAttribute("evaluatorType");
      if (evaluatorType.isEmpty())
        throw new InvalidModelException("evaluator node must have a non-empty \"evaluatorType\" attribute.");

      NodeList stressNodes = element.getElementsByTagName("stress");

      // dispatch on the type of evaluator:
      if (evaluatorType.equalsIgnoreCase("BoxabilityEvaluator"))
        return createBoxability(model, stressNodes);
      else if (evaluatorType.equalsIgnoreCase("ClimbEvaluator"))
        return createClimb(model, stressNodes);
      else if (evaluatorType.equalsIgnoreCase("CurrentValueEvaluator"))
        return createCurrentValue(model, stressNodes);
      else if (evaluatorType.equalsIgnoreCase("FlippingEvaluator"))
        return createFlipping(model, stressNodes);
      else if (evaluatorType.equalsIgnoreCase("MotorComponentViabilityEvaluator"))
        return createMotorComponent(model, stressNodes);
      // don't think we need the proto evaluator anymore, but the code is here if needed
      //else if (evaluatorType.equalsIgnoreCase("ProtoEvaluator"))
      //  return createProto(model, stressNodes);
      else if (evaluatorType.equalsIgnoreCase("RobotDifferenceEvaluator"))
        return createRobotDifference(model, dModel, stressNodes);
      else if (evaluatorType.equalsIgnoreCase("XMLRPCEvaluator"))
        return createXMLRPC(cModel, stressNodes);
      else
        throw new InvalidModelException("Unknown evaluator type: \""+evaluatorType+".\"");
    }
    else
      // if the functional blueprint is not active then we should just 
      // stop and return null
      return null;
  }

  private static Evaluator createBoxability(AttributeValueView model, 
      NodeList stressNodes) throws InvalidModelException {
    List<StressNodeData> data = parseStressNodes(stressNodes);

    if (data.size() != 1)
      throw new InvalidModelException("Expected exactly one \"stress\" node per boxability evaluator.");

    StressNodeData boxData = data.get(0);
    String[] args = boxData.getArgs();
    if(args.length != 3)
      throw new InvalidModelException("Expecting exactly 3 \"arg\" nodes for BoxabilityEvaluator.");

    return new BoxabilityEvaluator(model, boxData.getName(), boxData.getStressType(), 
        args[0], args[1], args[2]);
  }

  private static Evaluator createClimb(AttributeValueView model, 
      NodeList stressNodes) throws InvalidModelException {
    List<StressNodeData> data = parseStressNodes(stressNodes);

    if (data.size() != 1)
      throw new InvalidModelException("Expected exactly one \"stress\" node per climb evaluator.");

    StressNodeData climbData = data.get(0);
    String[] args = climbData.getArgs();
    if(args.length != 3)
      throw new InvalidModelException("Expecting exactly 3 \"arg\" nodes for ClimbEvaluator.");

    return new ClimbEvaluator(model, climbData.getName(), climbData.getStressType(), args[0], 
        args[1], args[2]);
  }  

  private static Evaluator createFlipping(AttributeValueView model, 
      NodeList stressNodes) throws InvalidModelException {
    List<StressNodeData> data = parseStressNodes(stressNodes);

    if (data.size() != 1)
      throw new InvalidModelException("Expected exactly one \"stress\" node per climb evaluator.");

    StressNodeData flipData = data.get(0);
    String[] args = flipData.getArgs();
    if(args.length != 4)
      throw new InvalidModelException("Expecting exactly 4 \"arg\" nodes for FlippingEvaluator.");

    return new FlippingEvaluator(model, flipData.getName(), flipData.getStressType(), args[0], 
        args[1], args[2], args[3]);
  }  

  private static Evaluator createCurrentValue(AttributeValueView model, 
      NodeList stressNodes) throws InvalidModelException {
    List<StressNodeData> data = parseStressNodes(stressNodes);

    List<String> stressTypes = new ArrayList<String>();
    List<String> stressFeatures = new ArrayList<String>();
    List<String> stressInputs = new ArrayList<String>();

    for (StressNodeData cv : data) {
      String[] args = cv.getArgs();
      if(args.length != 1)
        throw new InvalidModelException("Expecting exactly 1 \"arg\" node for CurrentValueEvaluator.");

      stressTypes.add(cv.getStressType());
      stressFeatures.add(args[0]);
      stressInputs.add(cv.getName());
      cv.getName();
      cv.getStressType();
      cv.getArgs();
    }

    return new CurrentValueEvaluator(model, stressTypes, stressFeatures, stressInputs);
  }

  private static Evaluator createMotorComponent(AttributeValueView model, 
      NodeList stressNodes) throws InvalidModelException {
    List<StressNodeData> data = parseStressNodes(stressNodes);

    if (data.size() != 1)
      throw new InvalidModelException("Expected exactly one \"stress\" node per climb evaluator.");

    StressNodeData mcData = data.get(0);
    String[] args = mcData.getArgs();
    if(args.length != 2)
      throw new InvalidModelException("Expecting exactly 2 \"arg\" nodes for MotorComponentViabilityEvaluator.");

    return new MotorComponentViabilityEvaluator(model, mcData.getName(), 
        mcData.getStressType(), args[0], args[1]);
  }

  // don't think we need the proto evaluator, but the code is here if needed:
  //  private static Evaluator createProto(CompositeModel model, 
  //      NodeList stressNodes) throws InvalidModelException {
  //    List<StressNodeData> data = parseStressNodes(stressNodes);
  //    
  //    if (data.size() != 1)
  //      throw new InvalidModelException("Expected exactly one \"stress\" node per proto evaluator.");
  //
  //    StressNodeData protoData = data.get(0);
  //    String[] args = protoData.getArgs();
  //    if(args.length != 0)
  //      throw new InvalidModelException("Expecting exactly 0 \"arg\" nodes for ProtoEvaluator.");
  //    
  //    return new ProtoEvaluator(model, protoData.getName(), protoData.getStressType());
  //  }
  // needed lines for xml file:
  //  <evaluator evaluatorType="ProtoEvaluator" active="true">
  //  <stress name="Climb FB" type="angle" args="0"/>
  //  </evaluator>

  private static Evaluator createRobotDifference(AttributeValueView avv, 
      DesiredAttributeValueView desired, NodeList stressNodes) 
          throws InvalidModelException {
    List<StressNodeData> data = parseStressNodes(stressNodes);

    if (data.size() != 1)
      throw new InvalidModelException("Expected exactly one \"stress\" node per robot difference evaluator.");

    StressNodeData diffData = data.get(0);
    String[] args = diffData.getArgs();
    if(args.length != 0)
      throw new InvalidModelException("Expecting exactly 0 \"arg\" nodes for RobotDifferenceEvaluator.");

    return new RobotDifferenceEvaluator(avv, desired, diffData.getStressType());
  }

    private static Evaluator createXMLRPC(CompositeModel model, 
        NodeList stressNodes) throws InvalidModelException {
      List<StressNodeData> data = parseStressNodes(stressNodes);
      
      if (data.size() != 1)
        throw new InvalidModelException("Expected exactly one \"stress\" node per xmlrpc evaluator.");
  
      StressNodeData xmlrpcData = data.get(0);
      String[] args = xmlrpcData.getArgs();
      if(args.length != 0)
        throw new InvalidModelException("Expecting exactly 0 \"arg\" nodes for XMLRPCEvaluator.");
      
      return new XMLRPCEvaluator(model, xmlrpcData.getName(), xmlrpcData.getStressType());
    }

  // generalize the parsing of stress nodes:
  private static List<StressNodeData> parseStressNodes(NodeList stressNodes) 
      throws InvalidModelException {
    List<StressNodeData> data = new ArrayList<StressNodeData>();

    for (int i=0; i<stressNodes.getLength(); i++) {
      Node node = stressNodes.item(i);
      if (node.getNodeType() != Node.ELEMENT_NODE) 
        throw new InvalidModelException("\"stress\" node not Element as expected");

      Element stressElement = (Element) node;
      String name = stressElement.getAttribute("name");
      if (name.isEmpty())
        throw new InvalidModelException("stress node must have a non-empty \"name\" attribute.");

      String stressType = stressElement.getAttribute("type");
      if (stressType.isEmpty())
        throw new InvalidModelException("stress node must have a non-empty \"type\" attribute.");

      String argsStr = stressElement.getAttribute("args");
      if (argsStr.isEmpty())
        throw new InvalidModelException("stress node must have a non-empty \"args\" attribute.");
      int numArgs = Integer.parseInt(argsStr);
      String[] args = getArgs(numArgs, stressElement);

      data.add(new StressNodeData(name, stressType, args));
    }
    return data;
  }

  private static String[] getArgs(int numArgs, Element parentElement) throws InvalidModelException {
    NodeList argNodes = parentElement.getElementsByTagName("arg");
    if (argNodes.getLength() != numArgs)
      throw new InvalidModelException("Number of \"arg\" nodes does not match the specified number of arguments");

    String[] args = new String[numArgs];

    for (int i=0; i<numArgs; i++) {
      Node argNode = argNodes.item(i);
      if (argNode.getNodeType() != Node.ELEMENT_NODE) 
        throw new InvalidModelException("\"arg\" node not Element as expected");
      Element argE = (Element)argNode;
      String numStr = argE.getAttribute("num");
      if (numStr.isEmpty())
        throw new InvalidModelException("\"arg\" node must have a non-empty \"num\" attribute.");
      int num = Integer.parseInt(numStr);
      if ((num <= 0) || (num > numArgs))
        throw new InvalidModelException("\"arg\" node must have a \"num\" value between 1 and "+numArgs+" and has a value of: "+num+".");
      if (args[num-1] != null)
        throw new InvalidModelException("multiple \"arg\" nodes have the same \"num\" attribute.");
      else {
        String param = argE.getTextContent();
        args[num-1] = param;
      }
    }

    return args;
  }

  private static class StressNodeData {
    String m_name;
    String m_stressType;
    String[] m_args;

    public StressNodeData(String name, String stressType, String[] args) {
      m_name = name;
      m_stressType = stressType;
      m_args = args;
    }

    public String getName() {
      return m_name;
    }

    public String getStressType() {
      return m_stressType;
    }

    public String[] getArgs() {
      return m_args;
    }
  }

}
