/* Parsing tests for evaluators
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */

package com.bbn.morph.evaluator.test;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.List;

import junit.framework.Assert;

import org.junit.Before;
import org.junit.Test;

import com.bbn.morph.evaluator.BoxabilityEvaluator;
import com.bbn.morph.evaluator.ClimbEvaluator;
import com.bbn.morph.evaluator.CurrentValueEvaluator;
import com.bbn.morph.evaluator.Evaluator;
import com.bbn.morph.evaluator.EvaluatorParser;
import com.bbn.morph.evaluator.FlippingEvaluator;
import com.bbn.morph.evaluator.MotorComponentViabilityEvaluator;
import com.bbn.morph.evaluator.RobotDifferenceEvaluator;
import com.bbn.morph.evaluator.XMLRPCEvaluator;
import com.bbn.morph.model.CompositeModel;
import com.bbn.morph.model.InvalidModelException;
import com.bbn.morph.view.AttributeValueView;
import com.bbn.morph.view.DesiredAttributeValueView;

public class EvaluatorParserTest {
  private CompositeModel m_compositeModel;
  private AttributeValueView m_attVV;
  private DesiredAttributeValueView m_desired;

  
  /**
   * Sets up the test fixture. 
   * (Called before every test case method.)
   */
  @Before
  public void setUp() {
    // for the test we can get away with setting these to null
    m_compositeModel = null;
    m_attVV = null;
    m_desired = null;
  }
  
  /**
   * Test that inactive evaluators are not processed.
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test
  public void inactiveTest() throws IOException, InvalidModelException {
    // test with active set to false
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\" active=\"false\"/></evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    List<Evaluator> evaluators = EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
    Assert.assertTrue(evaluators.isEmpty());
  }
  
  
  /**
   * Test for root node name
   * @throws InvalidModelException 
   */
  @Test(expected=InvalidModelException.class)
  public void rootNodeTest() throws IOException, InvalidModelException {
    String s = "<evaluator/>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for evaluator node name
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeTest() throws IOException, InvalidModelException {
    String s = "<evaluators><eval/></evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for evaluator evaluatorType attribute
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeEvaluatorTypeTest() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator/></evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for evaluator evaluatorType attribute as something invalid
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeEvaluatorTypeTest2() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator evaluatorType=\"foo\"/></evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for BoxabilityEvaluator, test 1
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeBoxabilityEvaluatorTest1() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\"/></evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for BoxabilityEvaluator, test 2
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeBoxabilityEvaluatorTest2() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\">";
    s+="<stress/></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for BoxabilityEvaluator, test 3
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeBoxabilityEvaluatorTest3() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\">";
    s+="<stress name=\"test\"/></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for BoxabilityEvaluator, test 4
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeBoxabilityEvaluatorTest4() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\"/></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for BoxabilityEvaluator, test 5
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeBoxabilityEvaluatorTest5() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"0\"/></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for BoxabilityEvaluator, test 6
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeBoxabilityEvaluatorTest6() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"3\"/></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for BoxabilityEvaluator, test 7
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeBoxabilityEvaluatorTest7() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"foo\"/></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for BoxabilityEvaluator, test 8
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeBoxabilityEvaluatorTest8() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"3\">";
    s+="<arg/>";
    s+="<arg/>";
    s+="<arg/>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for BoxabilityEvaluator, test 9
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeBoxabilityEvaluatorTest9() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"3\">";
    s+="<arg num=\"1\"/>";
    s+="<arg num=\"2\"/>";
    s+="<arg num=\"1\"/>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for BoxabilityEvaluator, test 10
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeBoxabilityEvaluatorTest10() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"3\">";
    s+="<arg num=\"1\"/>";
    s+="<arg num=\"2\"/>";
    s+="<arg num=\"4\"/>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for BoxabilityEvaluator, test 11
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeBoxabilityEvaluatorTest11() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"3\">";
    s+="<arg num=\"1\"/>";
    s+="<arg num=\"2\"/>";
    s+="<arg num=\"0\"/>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for BoxabilityEvaluator, test 12
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeBoxabilityEvaluatorTest12() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"3\">";
    s+="<arg num=\"1\"/>";
    s+="<arg num=\"2\"/>";
    s+="<arg num=\"-5\"/>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for BoxabilityEvaluator, test 13
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeBoxabilityEvaluatorTest13() throws IOException, InvalidModelException {
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"2\">";
    s+="<arg num=\"1\"/>";
    s+="<arg num=\"2\"/>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for BoxabilityEvaluator, test 14, should work
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test
  public void evaluatorNodeBoxabilityEvaluatorTest14() throws Exception {
    String s = "<evaluators><evaluator evaluatorType=\"BoxabilityEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"3\">";
    s+="<arg num=\"1\">arg1</arg>";
    s+="<arg num=\"2\">arg2</arg>";
    s+="<arg num=\"3\">arg3</arg>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    List<Evaluator> evaluators = EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
    Assert.assertEquals(1, evaluators.size());
    
    Evaluator evaluator = evaluators.get(0);
    Assert.assertTrue(evaluator instanceof BoxabilityEvaluator);
    BoxabilityEvaluator be = (BoxabilityEvaluator)evaluator;
    
    // now need to make sure that the BoxabilityEvaluator is okay
    Field field = be.getClass().getDeclaredField("m_rFlip");
    field.setAccessible(true);
    String str = (String)field.get(be);
    Assert.assertEquals("arg1", str);
    
    field = be.getClass().getDeclaredField("m_lFlip");
    field.setAccessible(true);
    str = (String)field.get(be);
    Assert.assertEquals("arg2", str);
    
    field = be.getClass().getDeclaredField("m_body");
    field.setAccessible(true);
    str = (String)field.get(be);
    Assert.assertEquals("arg3", str);
    
    field = be.getClass().getDeclaredField("m_featureName");
    field.setAccessible(true);
    str = (String)field.get(be);
    Assert.assertEquals("test", str);
    
    field = be.getClass().getDeclaredField("m_stressType");
    field.setAccessible(true);
    str = (String)field.get(be);
    Assert.assertEquals("alpha", str);
  }
  
  /**
   * Test for ClimbEvaluator, test 1
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeClimbEvaluatorTest1() throws InvalidModelException, IOException {
    String s = "<evaluators><evaluator evaluatorType=\"ClimbEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"2\">";
    s+="<arg num=\"1\">arg1</arg>";
    s+="<arg num=\"2\">arg2</arg>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for ClimbEvaluator, test 2, should work
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test
  public void evaluatorNodeClimbEvaluatorTest2() throws Exception {
    String s = "<evaluators><evaluator evaluatorType=\"ClimbEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"3\">";
    s+="<arg num=\"1\">arg1</arg>";
    s+="<arg num=\"2\">arg2</arg>";
    s+="<arg num=\"3\">arg3</arg>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    List<Evaluator> evaluators = EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
    Assert.assertEquals(1, evaluators.size());
    
    Evaluator evaluator = evaluators.get(0);
    Assert.assertTrue(evaluator instanceof ClimbEvaluator);
    ClimbEvaluator ce = (ClimbEvaluator)evaluator;
    
    // now need to make sure that the ClimbEvaluator is okay
    Field field = ce.getClass().getDeclaredField("m_rFlip");
    field.setAccessible(true);
    String str = (String)field.get(ce);
    Assert.assertEquals("arg1", str);
    
    field = ce.getClass().getDeclaredField("m_lFlip");
    field.setAccessible(true);
    str = (String)field.get(ce);
    Assert.assertEquals("arg2", str);
    
    field = ce.getClass().getDeclaredField("m_body");
    field.setAccessible(true);
    str = (String)field.get(ce);
    Assert.assertEquals("arg3", str);
    
    field = ce.getClass().getDeclaredField("m_featureName");
    field.setAccessible(true);
    str = (String)field.get(ce);
    Assert.assertEquals("test", str);
    
    field = ce.getClass().getDeclaredField("m_stressType");
    field.setAccessible(true);
    str = (String)field.get(ce);
    Assert.assertEquals("alpha", str);
  }
  
  /**
   * Test for FlippingEvaluator, test 1
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeFlippingEvaluatorTest1() throws InvalidModelException, IOException {
    String s = "<evaluators><evaluator evaluatorType=\"FlippingEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"2\">";
    s+="<arg num=\"1\">arg1</arg>";
    s+="<arg num=\"2\">arg2</arg>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for FlippingEvaluator, test 2, should work
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test
  public void evaluatorNodeFlippingEvaluatorTest2() throws Exception {
    String s = "<evaluators><evaluator evaluatorType=\"FlippingEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"4\">";
    s+="<arg num=\"1\">arg1</arg>";
    s+="<arg num=\"2\">arg2</arg>";
    s+="<arg num=\"3\">arg3</arg>";
    s+="<arg num=\"4\">arg4</arg>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    List<Evaluator> evaluators = EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
    Assert.assertEquals(1, evaluators.size());
    
    Evaluator evaluator = evaluators.get(0);
    Assert.assertTrue(evaluator instanceof FlippingEvaluator);
    FlippingEvaluator fe = (FlippingEvaluator)evaluator;
    
    // now need to make sure that the FlippingEvaluator is okay
    Field field = fe.getClass().getDeclaredField("m_rFlip");
    field.setAccessible(true);
    String str = (String)field.get(fe);
    Assert.assertEquals("arg1", str);
    
    field = fe.getClass().getDeclaredField("m_lFlip");
    field.setAccessible(true);
    str = (String)field.get(fe);
    Assert.assertEquals("arg2", str);
    
    field = fe.getClass().getDeclaredField("m_body");
    field.setAccessible(true);
    str = (String)field.get(fe);
    Assert.assertEquals("arg3", str);

    field = fe.getClass().getDeclaredField("m_torque");
    field.setAccessible(true);
    str = (String)field.get(fe);
    Assert.assertEquals("arg4", str);
    
    field = fe.getClass().getDeclaredField("m_featureName");
    field.setAccessible(true);
    str = (String)field.get(fe);
    Assert.assertEquals("test", str);
    
    field = fe.getClass().getDeclaredField("m_stressType");
    field.setAccessible(true);
    str = (String)field.get(fe);
    Assert.assertEquals("alpha", str);
  }
  
  /**
   * Test for MotorComponentViabilityEvaluator, test 1
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeMotorComponentViabilityEvaluatorTest1() throws InvalidModelException, IOException {
    String s = "<evaluators><evaluator evaluatorType=\"MotorComponentViabilityEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"1\">";
    s+="<arg num=\"1\">arg1</arg>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for MotorComponentViabilityEvaluator, test 2, should work
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test
  public void evaluatorNodeMotorComponentViabilityEvaluatorTest2() throws Exception {
    String s = "<evaluators><evaluator evaluatorType=\"MotorComponentViabilityEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"2\">";
    s+="<arg num=\"1\">arg1</arg>";
    s+="<arg num=\"2\">arg2</arg>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    List<Evaluator> evaluators = EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
    Assert.assertEquals(1, evaluators.size());
    
    Evaluator evaluator = evaluators.get(0);
    Assert.assertTrue(evaluator instanceof MotorComponentViabilityEvaluator);
    MotorComponentViabilityEvaluator me = (MotorComponentViabilityEvaluator)evaluator;
    
    // now need to make sure that the MotorComponentViabilityEvaluator is okay
    Field field = me.getClass().getDeclaredField("m_motorMass");
    field.setAccessible(true);
    String str = (String)field.get(me);
    Assert.assertEquals("arg1", str);
    
    field = me.getClass().getDeclaredField("m_torque");
    field.setAccessible(true);
    str = (String)field.get(me);
    Assert.assertEquals("arg2", str);
    
    field = me.getClass().getDeclaredField("m_featureName");
    field.setAccessible(true);
    str = (String)field.get(me);
    Assert.assertEquals("test", str);
    
    field = me.getClass().getDeclaredField("m_stressType");
    field.setAccessible(true);
    str = (String)field.get(me);
    Assert.assertEquals("alpha", str);
  }
  
  
  /**
   * Test for RobotDifferenceEvaluator, test 1
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeRobotDifferenceEvaluatorTest1() throws InvalidModelException, IOException {
    String s = "<evaluators><evaluator evaluatorType=\"RobotDifferenceEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"1\">";
    s+="<arg num=\"1\">arg1</arg>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for RobotDifferenceEvaluator, test 2, should work
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test
  public void evaluatorNodeRobotDifferenceEvaluatorTest2() throws Exception {
    String s = "<evaluators><evaluator evaluatorType=\"RobotDifferenceEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"0\">";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    List<Evaluator> evaluators = EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
    Assert.assertEquals(1, evaluators.size());
    
    Evaluator evaluator = evaluators.get(0);
    Assert.assertTrue(evaluator instanceof RobotDifferenceEvaluator);
    RobotDifferenceEvaluator me = (RobotDifferenceEvaluator)evaluator;
    
    // now need to make sure that the RobotDifferenceEvaluator is okay
    
    Field field = me.getClass().getDeclaredField("m_stressType");
    field.setAccessible(true);
    String str = (String)field.get(me);
    Assert.assertEquals("alpha", str);
  }
  
  /**
   * Test for XMLRPCEvaluator, test 1
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeXMLRPCEvaluatorTest1() throws InvalidModelException, IOException {
    String s = "<evaluators><evaluator evaluatorType=\"XMLRPCEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"1\">";
    s+="<arg num=\"1\">arg1</arg>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for XMLRPCEvaluator, test 2, should work
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test
  public void evaluatorNodeXMLRPCEvaluatorTest2() throws Exception {
    String s = "<evaluators><evaluator evaluatorType=\"XMLRPCEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"0\">";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    List<Evaluator> evaluators = EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
    Assert.assertEquals(1, evaluators.size());
    
    Evaluator evaluator = evaluators.get(0);
    Assert.assertTrue(evaluator instanceof XMLRPCEvaluator);
    XMLRPCEvaluator me = (XMLRPCEvaluator)evaluator;
    
    // now need to make sure that the RobotDifferenceEvaluator is okay
    
    Field field = me.getClass().getDeclaredField("m_featureName");
    field.setAccessible(true);
    String str = (String)field.get(me);
    Assert.assertEquals("test", str);
    
    field = me.getClass().getDeclaredField("m_stressType");
    field.setAccessible(true);
    str = (String)field.get(me);
    Assert.assertEquals("alpha", str);
  }
  
  
  /**
   * Test for CurrentValueEvaluator, test 1
   * @throws IOException
   * @throws InvalidModelException
   */
  @Test(expected=InvalidModelException.class)
  public void evaluatorNodeCurrentValueEvaluatorTest1() throws InvalidModelException, IOException {
    String s = "<evaluators><evaluator evaluatorType=\"CurrentValueEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"2\">";
    s+="<arg num=\"1\">arg1</arg>";
    s+="<arg num=\"2\">arg2</arg>";
    s+="</stress></evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
  }
  
  /**
   * Test for CurrentValueEvaluator, test 2, should work
   * @throws IOException
   * @throws InvalidModelException
   */
  @SuppressWarnings("unchecked")
  @Test
  public void evaluatorCurrentValueEvaluatorTest2() throws Exception {
    String s = "<evaluators><evaluator evaluatorType=\"CurrentValueEvaluator\">";
    s+="<stress name=\"test\" type=\"alpha\" args=\"1\">";
    s+="<arg num=\"1\">arg1</arg>";
    s+="</stress>";
    s+="<stress name=\"test2\" type=\"beta\" args=\"1\">";
    s+="<arg num=\"1\">arg1</arg>";
    s+="</stress>";
    s+="</evaluator>";
    s+="</evaluators>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    List<Evaluator> evaluators = EvaluatorParser.parseEvaluators(stream, m_compositeModel, m_attVV, m_desired);
    Assert.assertEquals(1, evaluators.size());
    
    Evaluator evaluator = evaluators.get(0);
    Assert.assertTrue(evaluator instanceof CurrentValueEvaluator);
    CurrentValueEvaluator fe = (CurrentValueEvaluator)evaluator;
    
    // now need to make sure that the CurrentValueEvaluator is okay
    Field field = fe.getClass().getDeclaredField("m_stressTypes");
    field.setAccessible(true);
    List<String> strs = (List<String>)field.get(fe);
    Assert.assertEquals("alpha", strs.get(0));
    Assert.assertEquals("beta", strs.get(1));
    
    field = fe.getClass().getDeclaredField("m_stressFeatures");
    field.setAccessible(true);
    strs = (List<String>)field.get(fe);
    Assert.assertEquals("arg1", strs.get(0));
    Assert.assertEquals("arg1", strs.get(1));

    field = fe.getClass().getDeclaredField("m_stressInputs");
    field.setAccessible(true);
    strs = (List<String>)field.get(fe);
    Assert.assertEquals("test", strs.get(0));
    Assert.assertEquals("test2", strs.get(1));
  }
}
