/** Functional Blueprint unit tests
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. 
**/

package com.bbn.morph.executer.test;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import junit.framework.Assert;

import org.junit.Test;

import com.bbn.morph.view.FB;
import com.bbn.morph.executer.FBExecutor;
import com.bbn.morph.executer.FBModel;
import com.bbn.morph.executer.Feature;
import com.bbn.morph.executer.Pair;
import com.bbn.morph.executer.Update;
import com.bbn.morph.view.ThreePieceLinearStressFunction;
import com.bbn.morph.view.RobotAttribute;
import com.bbn.morph.view.AttributeValueView;
import com.bbn.morph.view.StepDecrementUpdate;
import com.bbn.morph.view.StepIncrementUpdate;
import com.bbn.morph.view.StressFunction;
import com.bbn.morph.executer.StressInputType;
import com.bbn.morph.executer.StressInputs;
import com.bbn.morph.model.Measurement;
import com.bbn.morph.model.Units;
import com.bbn.morph.view.UpdateFunction;

public class FBExecutorTest {
	final static String attribute1 = "FlipperLength";
	final static String attribute2 = "BodyLength";
	final static String feature1 = "ClimbFB";
	final static String feature2 = "BoxFB";
	
  static  public AttributeValueView buildRobotModel(){
    	AttributeValueView rm= new AttributeValueView(null);
    	rm.put(new RobotAttribute(attribute1), new Measurement(Units.getMeters(), 10.0));
    	rm.put(new RobotAttribute(attribute2), new Measurement(Units.getMeters(), 20.0));
    	return rm;	
    }
    // FB for climbing
   static  public FB createFBfeature1(){
		Pair first =new Pair(3000,0);
		Pair second =new Pair(5000,10000);
		StressInputType inputLabel= StressInputType.TIME;
		StressFunction stressF= new ThreePieceLinearStressFunction(inputLabel, first, second);
		UpdateFunction updateF1= new StepDecrementUpdate(0.5);
		UpdateFunction updateF2= new StepIncrementUpdate(0.6);
		HashMap<RobotAttribute, UpdateFunction> updateFunctions = new HashMap<RobotAttribute, UpdateFunction>();
		updateFunctions.put(new RobotAttribute(attribute1), updateF1);
		updateFunctions.put(new RobotAttribute(attribute2), updateF2);
		FB testFB = new FB(new Feature(feature1), updateFunctions, stressF);
		return testFB;
	}
    
 // FB for Mech Stress
   static  public FB createFBfeature2(){
		Pair first =new Pair(4,0);
		Pair second =new Pair(8,20);
		StressInputType inputLabel= StressInputType.WEIGHT;
		StressFunction stressF= new ThreePieceLinearStressFunction(inputLabel, first, second);
		UpdateFunction updateF1= new StepIncrementUpdate(0.3);
		UpdateFunction updateF2= new StepIncrementUpdate(0.8);
		HashMap<RobotAttribute, UpdateFunction> updateFunctions = new HashMap<RobotAttribute, UpdateFunction>();
		updateFunctions.put(new RobotAttribute(attribute1), updateF1);
		updateFunctions.put(new RobotAttribute(attribute2), updateF2);
		FB testFB = new FB(new Feature(feature2), updateFunctions, stressF);
		return testFB;
	}
   static public FBExecutor createFBExecutor(){
	   FBModel fbModel= new FBModel();
	   fbModel.put(new Feature(feature1), createFBfeature1());
	   fbModel.put(new Feature(feature2), createFBfeature2());
	   FBExecutor fbExecutor = new FBExecutor(buildRobotModel(), fbModel);
	   return fbExecutor;
	   
	   
   }
	static public StressInputs genInput(String FBname, StressInputType inputLabel, double val){
		StressInputs inputs = new StressInputs(new Feature(FBname));
		inputs.put(inputLabel, val);
		return inputs;
	}
@Test
	public void testFBExecutor(){
		FBExecutor fbExecutor = createFBExecutor();
		Set<StressInputs> stressInputSet= new HashSet<StressInputs>();
		stressInputSet.add(genInput(feature1,StressInputType.TIME, 4000));
		stressInputSet.add(genInput(feature2,StressInputType.WEIGHT, 5));
		
		Map<FB, Double >  fbStress = fbExecutor.compoundStress(stressInputSet);
		System.out.println("Stress:"+ fbStress);
		// Compute the suggested updates by each FB
		Map<RobotAttribute, Set<Update>> singleFBUpdates = fbExecutor.fbUpdates(fbStress, 1.0);
		System.out.println("Single Update:"+ singleFBUpdates);
		// Combine the suggestions and produce one update per FB
		Set<Update>  compositeFBUpdate= fbExecutor.compoundUpdate(singleFBUpdates);
		System.out.println("Compound Update:"+ compositeFBUpdate);
		
		
		boolean changed = fbExecutor.localRun(stressInputSet);
		if (changed)
		  System.out.println("robot model updated");
		else
		  System.out.println("robot model not updated");
		System.out.println(fbExecutor.getRobotModel());
	}


@Test
   public void testBlend(){
	List<Double> values = new LinkedList<Double>();
	values.add(3.0);
	// just one input: {3}
	double  result = FBExecutor.blend(values);
	Assert.assertEquals(result, 3.0);
	
	// {3 ,3}
	values.add(new Double(3.0));
    result = FBExecutor.blend(values);
	Assert.assertEquals(result, 3.0);
	
	
	// {3 ,3, 5}
	values.add(5.0);
    result = FBExecutor.blend(values);
	Assert.assertEquals(result, 5.0);
	
	// just one input {-3}
	values.clear();
	values.add(-3.0);
    result = FBExecutor.blend(values);
	Assert.assertEquals(result, -3.0);
	// {-3 ,-3}
	values.add(-3.0);
    result = FBExecutor.blend(values);
	Assert.assertEquals(result, -3.0);
	
	// {-3 ,-3, -5}
	values.add(-5.0);
    result = FBExecutor.blend(values);
	Assert.assertEquals(result, -5.0);
	
	// two equal strength opposite side
	// {-3, 3}
	values.clear();
	values.add(3.0);
	values.add(-3.0);
    result = FBExecutor.blend(values);
	Assert.assertEquals(result, 0.0);
	
	// two unequal strength opposite side
	// {-3, 5}
	values.clear();
	values.add(5.0);
	values.add(-3.0);
    result = FBExecutor.blend(values);
	Assert.assertEquals(result, 10.0/8.0);
	
	// multiple unequal strength opposite side
	// {-3, -4 , -2, 5, 5}
	values.clear();
	values.add(5.0);
	values.add(5.0);
	values.add(-3.0);
	values.add(-4.0);
	values.add(-2.0);
    result = FBExecutor.blend(values);
    Assert.assertEquals(result, 5.0/19.0);
}
public static  void main(String[] args) {
	
}
}

