/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */

package com.bbn.morph.model;

import java.util.HashSet;
import java.util.Set;



/**
 * The Measurement class stores measurable quantities with a value and an associated unit.
 * @author kmcguire
 * @author aadler
 */
public class Measurement{
	private Units m_units;
	private double m_value;
	
	/**
	 * stores any measurable quantity
	 * as if in MGS units (use getStandardValue to use MKS units)
	 * 
	 * @author kmcguire
	 * @param units takes {@link UnitNames} and {@link UnitPrefixes}
	 * @param value
	 */
	public Measurement(Units units, double value){
		m_units = units;
		m_value = value * units.getConvert();
	}
	// stores value as if in MGS units (getStandardValue forces computations to be made in MKS units)
	
	/**
	 * @return a copy of the Measurement
	 */
	public Measurement copy() {
	  return new Measurement(getUnits(), getPrefixValue());
	}
		
	/**
	 * Used to construct a measurement out of an input {@link String} from the UI.
	 * 
	 * 
	 * @author kmcguire
	 * @param str
	 * @param expectedUnit
	 * @return returns a measurement or null if the units don't match or there is no double.
	 */
	public static Measurement getMeasurement(String str, Units expectedUnit){
    str = str.trim();
    MeasurementParse parse = new MeasurementParse(str, expectedUnit.getPrefix(), expectedUnit.getBaseUnit());
    
    // we want to pull off the number at the front of the string:
    // we're not going to take negative numbers so we're expecting digits with at most one period

    StringBuffer sb = new StringBuffer();
    boolean hasPeriod = false;
    while (parse.getPosition() < parse.getString().length()) {
      char c = parse.getString().charAt(parse.getPosition());
      if (Character.isDigit(c)) {
        sb.append(c);
        parse.increasePosition(1);
      }
      else if (c == '.') {
        // one period allowed
        if (hasPeriod)
          // too many periods
          return null;
        else {
          hasPeriod = true;
          sb.append(c);
          parse.increasePosition(1);          
        }
      }
      else {
        // not a digit or period, so break
        break;
      }
    }
    
    // try to get the digit:
    try {
      double value = Double.parseDouble(sb.toString());


      // skip whitespace
      while ((parse.getPosition() < parse.getString().length()) &&
          (Character.isWhitespace(parse.getString().charAt(parse.getPosition()))))
        parse.increasePosition(1);

      // check to see if we have text left
      if (parse.getPosition() < parse.getString().length()) {
        Set<MeasurementParse> parsesWithPrefix = addPrefix(parse);
        Set<MeasurementParse> parsesWithUnits = new HashSet<MeasurementParse>();
        for (MeasurementParse p : parsesWithPrefix) {
          parsesWithUnits.addAll(addUnits(p));
        }

        Set<MeasurementParse> goodParses = new HashSet<MeasurementParse>();
        for (MeasurementParse p : parsesWithUnits) {
          // check to see if the resulting string is empty:
          if (p.getString().length() == p.getPosition())
            // check to make sure we match the expected units:
            if (expectedUnit.getBaseUnit() == p.getUnit())
              goodParses.add(p);
        }

        // so we should hopefully only have one match. If we don't have one, then we're in trouble
        if (goodParses.size() == 1) {
          MeasurementParse goodParse = goodParses.iterator().next();
          Units units = new Units(goodParse.getPrefix(), goodParse.getUnit());
          return new Measurement(units, value);
        }
        else
          // this is an error, so return null
          return null;
      }
      else {
        // no overriding units:
        return new Measurement(expectedUnit, value);
      }
	  }
    catch (NumberFormatException e) {
      // couldn't parse the double, so it's an error:
	    return null;
	  }
	}
	
	private static Set<MeasurementParse> addUnits(MeasurementParse p) {
	  Set<MeasurementParse> parses = new HashSet<MeasurementParse>();
	  
	  // one option is to have no units:
	  MeasurementParse copy = p.copy();
	  copy.setUnit(UnitNames.NONE);
	  parses.add(copy);
	  
	  String toMatch = p.getString().substring(p.getPosition());
	  
	  // otherwise go through all the possible unit names, names2, and symbols:
	  for (UnitNames possibleUnit : UnitNames.values()) {
	    if (toMatch.startsWith(possibleUnit.getSymbol())) {
	      copy = p.copy();
	      copy.setUnit(possibleUnit);
	      copy.increasePosition(possibleUnit.getSymbol().length());
	      parses.add(copy);
	    }
      if (toMatch.startsWith(possibleUnit.getName())) {
        copy = p.copy();
        copy.setUnit(possibleUnit);
        copy.increasePosition(possibleUnit.getName().length());
        parses.add(copy);
      }
      if (toMatch.startsWith(possibleUnit.getName2())) {
        copy = p.copy();
        copy.setUnit(possibleUnit);
        copy.increasePosition(possibleUnit.getName2().length());
        parses.add(copy);
      }
	  }
	  
	  return parses;
  }

  private static Set<MeasurementParse> addPrefix(MeasurementParse parse) {
    Set<MeasurementParse> parses = new HashSet<MeasurementParse>();
    
    // one option is to have no units:
    MeasurementParse copy = parse.copy();
    copy.setPrefix(UnitPrefixes.NONE);
    parses.add(copy);
    
    String toMatch = parse.getString().substring(parse.getPosition());
    
    // otherwise go through all the possible unit names, names2, and symbols:
    for (UnitPrefixes possiblePrefix : UnitPrefixes.values()) {
      if (toMatch.startsWith(possiblePrefix.getSymbol())) {
        copy = parse.copy();
        copy.setPrefix(possiblePrefix);
        copy.increasePosition(possiblePrefix.getSymbol().length());
        parses.add(copy);
      }
      if (toMatch.startsWith(possiblePrefix.getName())) {
        copy = parse.copy();
        copy.setPrefix(possiblePrefix);
        copy.increasePosition(possiblePrefix.getName().length());
        parses.add(copy);
      }
    }
    
    return parses;
  }
  
  public Units getUnits(){
		return m_units;
	}
	
	/**
	 * @return returns double value of the Measurement in the originally set prefix.
	 * @author kmcguire
	 */
	public double getPrefixValue(){
		return m_value / m_units.getConvert();
	}
	
	
	/**
	 * @param prefix
	 * @return returns a Measurement converted into the desired prefix value
	 * @author kmcguire
	 */
	public Measurement getValueIn(UnitPrefixes prefix){
		Units units = new Units(prefix, getUnits().getBaseUnit());
		double value = m_value / units.getConvert();
		return new Measurement(units, value);
	}
	
	/**
	 * @return returns a Measurement to the UI converted into more reasonable units depending on the size of its value.
	 * @author kmcguire
	 */
	public Measurement getNeaterMeasurement(){
		Measurement neat = this;
		double m = m_value;
		UnitPrefixes[] neatPrefixes = {UnitPrefixes.CENTI, UnitPrefixes.MILLI, UnitPrefixes.KILO, UnitPrefixes.NONE};
		//in order of last preference to first.
		for(UnitPrefixes d : neatPrefixes)
			if(Math.abs(m/d.getConvert())>=1 && Math.abs(m/d.getConvert())< 100)
				neat = getValueIn(d);
		return neat;
	}
	
	/**
	 * @return returns double value in standard MKS units.
	 */
	public double getStandardValue() {
	  if (getUnits().getBaseUnit() == UnitNames.GRAMS) {
	    return getValueIn(UnitPrefixes.KILO).getPrefixValue();
	  }
		else
			return m_value;
	}
	
	@Override
	public String toString() {
		return Double.toString(getStandardValue());
	}
		
	public Measurement add(Measurement m1){
		if(getUnits().getBaseUnit() != m1.getUnits().getBaseUnit())
			throw new MeasurementException("Cannot add Measurements of units " + getUnits().getBaseUnit() + " and " + m1.getUnits().getBaseUnit());
		Measurement m2 = new Measurement(getUnits().getStandardizedUnits(), getStandardValue()+ m1.getStandardValue());
		return m2;
	}
	
	public Measurement subtract(Measurement m1){
		if(getUnits().getBaseUnit() != m1.getUnits().getBaseUnit())
			throw new MeasurementException("Cannot subtract Measurements of units " + getUnits().getBaseUnit() + " and " + m1.getUnits().getBaseUnit());
		Measurement m2 = new Measurement(getUnits().getStandardizedUnits(), getStandardValue()- m1.getStandardValue());
		return m2;
	}
	
	public Measurement mult(Measurement m1){
		Measurement m2 = new Measurement(getUnits().getStandardizedUnits(), getStandardValue()* m1.getStandardValue());
		return m2;
	}
		
	public Measurement divide(Measurement m1){
		Measurement m2 = new Measurement(getUnits().getStandardizedUnits(), getStandardValue()/ m1.getStandardValue());
		return m2;
	}
		
	/**
	 * addMD() adds a double to the value of the measurement without changing it's units.
	 * @param d1
	 * @return Measurement
	 * @author kmcguire
	 */
	public Measurement addMD(double d1){
		return (new Measurement(getUnits().getStandardizedUnits(), getStandardValue() + d1)).getValueIn(getUnits().getPrefix());
	}
	
	/**
	 * subtractMD() subtracts a double from the value of a measurement and returns the measurement in the same units.
	 * @param d1
	 * @return Measurement
	 * @author kmcguire
	 */
	public Measurement subtractMD(double d1){
		return (new Measurement(getUnits().getStandardizedUnits(), getStandardValue() - d1)).getValueIn(getUnits().getPrefix());
	}
	
	/**
	 * multMD() multiplies a double and the value of a measurement, and returns a measurement with the same units.
	 * @param d1
	 * @return Measurement
	 * @author kmcguire
	 */
	public Measurement multMD(double d1){
		return (new Measurement(getUnits().getStandardizedUnits(), getStandardValue()*d1)).getValueIn(getUnits().getPrefix());
	}
	
	/**
	 * divideMD() divides the value of a measurement by a double, and returns a measurement with the same units.
	 * @param d1
	 * @return Measurement
	 * @author kmcguire
	 */
	public Measurement divideMD(double d1){
		return (new Measurement(getUnits().getStandardizedUnits(), getStandardValue()/d1)).getValueIn(getUnits().getPrefix());
		}
		
	public Measurement sqrt(){
		return (new Measurement(getUnits().getStandardizedUnits(), Math.sqrt(getStandardValue()))).getValueIn(getUnits().getPrefix());
	}
	
	public static Measurement max(Measurement m1, Measurement m2){
		return new Measurement(m1.getUnits().getStandardizedUnits(), Math.max(m1.getStandardValue(), m2.getStandardValue()));
	}
		
	public static Measurement min(Measurement m1, Measurement m2){
		return new Measurement(m1.getUnits().getStandardizedUnits(), Math.min(m1.getStandardValue(), m2.getStandardValue()));
	}
	}