/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */

package com.bbn.morph.model;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;


import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * An abstract class with subclasses representing the various types of objects
 * in the world: {@link ModelBox}, {@link ModelCylinder}, {@link ModelSphere}
 * @author aadler
 *
 */
public abstract class ModelBody extends ModelObject {
  
  private String m_id;
  private ModelPosition m_position;
  private ModelRotation m_rotation;
  private Measurement m_mass; // FIXME where does mass go? Not in xsd schema.
  
  // keep track of any extra parameters
  private Map<String, RobotAttributeValues> m_extraParameters = new HashMap<String, RobotAttributeValues>();
  
  // component links:
  private String m_libraryName = "";
  private ComponentLibrary m_library = null;
  private ComponentInstance m_component = null;
  
  // TODO do we need an index here? Maybe in a subclass?
  
  /**
   * Create a {@link ModelBody} with the specified parameters
   * @param id the unique identifier
   * @param libraryName the library name for this component or an empty string if none
   * @param position the position
   * @param rotation the rotation of the body, can be null
   * @param mass the mass
   */
  public ModelBody(String id, String libraryName, ModelPosition position, ModelRotation rotation, Measurement mass) {
    m_id = id;
    m_position = position;
    m_rotation = rotation;
    m_mass = mass;
    m_libraryName = libraryName;
  }
  
  /**
   * @return the name of the library to link to, or an empty string if there is none
   */
  public String getLibraryName() {
    return m_libraryName;
  }
  
  /**
   * @param av the attribute/value pair to add to the {@link ModelBody}
   */
  public void addExtraParameter(RobotAttributeValues av) {
    m_extraParameters.put(av.getAttribute(), av.copy());
  }

  /**
   * @param avs the attribute/value pairs to add to the {@link ModelBody}
   */
  public void addExtraParameters(List<RobotAttributeValues> avs) {
    for (RobotAttributeValues av: avs)
      addExtraParameter(av);
  }
  
  /**
   * @return the extra parameters as robot attribute values (does not copy)
   */
  public List<RobotAttributeValues> getExtraParameters() {
    return new ArrayList<RobotAttributeValues>(m_extraParameters.values());
  }
  
  /**
   * Clone the object, making a copy, but not copying object parents
   */
  public abstract ModelBody clone();

  /**
   * @return the mass of the body
   */
  public Measurement getMass() {
    return m_mass;
  }
  
  /**
   * @return the id of this {@link ModelBody}
   */
  public String getId() {
    return m_id;
  }
  
  /**
   * @return the position for this {@link ModelBody}
   */
  public ModelPosition getPosition() {
	  return m_position;
  }
  
  /**
   * Sets the body's position in space
   * @param position is the new position
   */
  public void setPosition(ModelPosition position){
	  m_position = position;
  }
  
  /**
   * @return the rotation of the body, may be null
   */
  public ModelRotation getRotation() {
    return m_rotation;
  }
  
  /**
   * @return a set of attributes for this {@link ModelBody} representing all of the parameter values
   */
  public Set<RobotAttributeValues> getAttributeValues() {
    Set<RobotAttributeValues> rav = new HashSet<RobotAttributeValues>();
    
    rav.add(new RobotAttributeValues(getId()+".x", m_position.getxPos()));
    rav.add(new RobotAttributeValues(getId()+".y", m_position.getyPos()));
    rav.add(new RobotAttributeValues(getId()+".z", m_position.getzPos()));
    
    if (m_rotation != null) {
      rav.addAll(m_rotation.getAttributes(getId()));
    }
    
    rav.add(new RobotAttributeValues(getId()+".mass", m_mass));
    
    // add the extra parameters:
    for (String attributes : m_extraParameters.keySet()) {
      RobotAttributeValues av = m_extraParameters.get(attributes);
      rav.add(new RobotAttributeValues(getId()+"."+av.getAttribute(), av.getMeasurement()));
    }

    return rav;
  }
  
  /**
   * Lookup the value of the specified attribute
   * @param attribute the attribute to look up
   * @return the value of the specified attribute
   */
  public Measurement getAttributeValue(String attribute) {
    // handle what we can handle:
    if (attribute.equals("x"))
      return m_position.getxPos();
    else if (attribute.equals("y"))
      return m_position.getyPos();
    else if (attribute.equals("z"))
      return m_position.getzPos();
    else if (attribute.equals("mass"))
      return m_mass;
    else if ((m_rotation != null) && (m_rotation.isHandledAttribute(attribute)))
    {
        Measurement rotation = m_rotation.getAttributeValue(attribute);
    	return rotation;
    }
    else if (m_extraParameters.containsKey(attribute))
      return m_extraParameters.get(attribute).getMeasurement();
    else
      // couldn't handle it anywhere!
      throw new RuntimeException("unhandled attribute "+attribute);
  }

  /**
   * Update the {@link ModelBody} based on the specified attribute and deltas and joints
   * @param robotAttributeValues the attribute and deltas to update
   * @param joints a list of attached joints that can be used when changing size
   */
  public void updateRobotAttributeFromDeltas(RobotAttributeValues robotAttributeValues,
      List<ModelJoint> joints) {
    // handle what we can handle:
    String attribute = robotAttributeValues.getAttribute();
    double delta = robotAttributeValues.getMeasurement().getStandardValue();
    
    if (attribute.equals("x"))
     {
    	Measurement x_position = m_position.getxPos();
        m_position.setxPos(new Measurement(x_position.getUnits(), delta + x_position.getStandardValue()));
     }
    else if (attribute.equals("y"))
    {
    	Measurement y_position = m_position.getyPos();
        m_position.setyPos(new Measurement(y_position.getUnits(), delta + y_position.getStandardValue()));
    }
    else if (attribute.equals("z"))
    {
       Measurement z_position = m_position.getzPos();
       m_position.setzPos(new Measurement(z_position.getUnits(), delta + z_position.getStandardValue()));
    }
    else if (attribute.equals("mass"))
    {
      m_mass = new Measurement(m_mass.getUnits(), m_mass.getStandardValue()+delta);
    }
    else if ((m_rotation != null) && (m_rotation.isHandledAttribute(attribute)))
      m_rotation.updateAttributeByDelta(attribute, delta);
    else if (m_extraParameters.containsKey(attribute)) {
      Measurement old = m_extraParameters.get(attribute).getMeasurement();
      m_extraParameters.put(attribute, new RobotAttributeValues(attribute, new Measurement(old.getUnits(), delta + old.getStandardValue())));
    }
    else
      // couldn't handle it anywhere!
      throw new RuntimeException("unhandled attribute "+attribute);
  }
  
  /**
   * Calculate what direction the joints are from the current body position. This really 
   * shouldn't return a {@link ModelPosition}, but we want to return a triple of 
   * data. 
   * @param joints a list of joints
   * @return 0,0,0 if the list of joints is empty, otherwise if all the joints 
   * are in a particular direction from the current body position, returns 1 or -1 in 
   * that part of the triple. 
   */
  protected ModelPosition getJointsDirection(List<ModelJoint> joints) {
    // TODO return some other type, or maybe it's okay.
    Measurement dx = null;
    Measurement dy = null;
    Measurement dz = null;
    ModelPosition startPosition = getTransformedPosition();
    
    for (ModelJoint joint : joints) {
      if (joint instanceof ModelHingeJoint) {
        ModelHingeJoint mhj = (ModelHingeJoint)joint;
        Double[] anchorV = {mhj.getAnchorVx(), mhj.getAnchorVy(), mhj.getAnchorVz()};
        Double[] tAnchor = transformAnchor(joint, anchorV);
        
        // figure out which direction the joint is from the start position:
        Measurement xDir = new Measurement(Units.getRadians(), Math.signum(tAnchor[0]-startPosition.getxPos().getStandardValue()));
        if (dx == null)
          dx = xDir;
        else if (dx.getStandardValue() != xDir.getStandardValue())
          dx = new Measurement(dx.getUnits(), 0d); // conflict
        
        Measurement yDir = new Measurement(Units.getRadians(), Math.signum(tAnchor[1]-startPosition.getyPos().getStandardValue()));
        if (dy == null)
          dy = yDir;
        else if (dy.getStandardValue() != yDir.getStandardValue())
          dy = new Measurement(dy.getUnits(), 0d); // conflict
        
        Measurement zDir = new Measurement(Units.getRadians(), Math.signum(tAnchor[2]-startPosition.getzPos().getStandardValue()));
        if (dz == null)
          dz = zDir;
        else if (dx.getStandardValue() != xDir.getStandardValue())
          dz = new Measurement(dz.getUnits(), 0d); // conflict
      }
      else {
        // FIXME handle this better
        //System.err.println("Unhandled joint type!");
      }
    }
    
    return new ModelPosition((dx==null)?new Measurement(Units.getMeters(), 0d):dx, (dy==null)?new Measurement(Units.getMeters(), 0d):dy, (dz==null)?new Measurement(Units.getMeters(), 0d):dz);
  }
  
  private Double[] transformAnchor(ModelJoint joint, Double[] initial) {
    Double[] copy = Arrays.copyOf(initial, initial.length); 
    
    // If it has a parent, the parent is a transform and should be applied
    ModelObject parent = joint.getParentObject();
    while((parent != null) && (parent instanceof ModelTransform )) {
      ModelTransform transform = (ModelTransform) parent;
      copy[0] += transform.getXShift();
      copy[1] += transform.getXShift();
      copy[2] += transform.getXShift();
      parent = parent.getParentObject();
    }

    return copy;
  }
  
  /**
   * @return the ModelPosition with transforms applied
   */
  private ModelPosition getTransformedPosition() {
    ModelPosition position = getPosition().copy();
    
    // If it has a parent, the parent is a transform and should be applied
    ModelObject parent = getParentObject();
    while((parent != null) && (parent instanceof ModelTransform )) {
      ModelTransform transform = (ModelTransform) parent;
      position.setxPos(position.getxPos().addMD(transform.getXShift()));
      position.setyPos(position.getyPos().addMD(transform.getYShift()));
      position.setzPos(position.getzPos().addMD(transform.getZShift()));
      parent = parent.getParentObject();
    }

    return position;
  }
  
  /**
   * Creates a {@link ModelBody} from the given XML element
   * @param element the XML element representing a {@link ModelBody}
   * @return the {@link ModelBody} encoded by the XML
   * @throws InvalidModelException if the XML is ill-formed
   */
  public static ModelBody createFromXML(Element element) throws InvalidModelException {
    NodeList allElements = element.getElementsByTagName("*");
    Set<String> tagNames = new HashSet<String>();
    for (int i=0; i<allElements.getLength(); i++)
      tagNames.add(allElements.item(i).getNodeName());
      
    if (!element.hasAttribute("type"))
      throw new InvalidModelException("body element missing type attribute");
    
    String type = element.getAttribute("type");
    
    String libraryName = "";
    if (element.hasAttribute("library"))
      libraryName = element.getAttribute("library");
    
    NodeList idNodes = element.getElementsByTagName("id");
    if (idNodes.getLength() != 1)
      throw new InvalidModelException("Did not have one and only one id element");
    String id = idNodes.item(0).getTextContent();
    tagNames.remove("id");
    
    NodeList positionNodes = element.getElementsByTagName("position");
    if (positionNodes.getLength() != 1)
      throw new InvalidModelException("Did not have one and only one position element");
    ModelPosition position = ModelPosition.createFromXML((Element)positionNodes.item(0));
    tagNames.remove("position");
    
    // FIXME not in schema!
    NodeList massNodes = element.getElementsByTagName("mass");
    if (massNodes.getLength() != 1)
      throw new InvalidModelException("Did not have one and only one mass element");
    Measurement mass = new Measurement(Units.getKilograms(), Double.parseDouble(massNodes.item(0).getTextContent()));
    tagNames.remove("mass");
    
    // rotation -- FIXME should this deal with something other than quaternion?
    ModelRotation rotation = null;
    NodeList quaternionNodes = element.getElementsByTagName("quaternion");
    if (quaternionNodes.getLength() == 1)
      rotation = ModelQuaternion.createFromXML((Element)quaternionNodes.item(0));
    tagNames.remove("quaternion");
    
    ModelBody body = null;
    if (type.equals(ModelBox.TYPE_NAME)) {
      body = ModelBox.createFromXML(element, id, libraryName, position, rotation, mass, tagNames);
    }
    else if (type.equals(ModelCylinder.TYPE_NAME)) {
      body = ModelCylinder.createFromXML(element, id, libraryName, position, rotation, mass, tagNames);
    }
    else if (type.equals(ModelSphere.TYPE_NAME)) {
      body = ModelSphere.createFromXML(element, id, libraryName, position, rotation, mass, tagNames);
    }
    else 
      throw new RuntimeException("Unhandled type: "+type);
    
    // need to add extra stuff to the body
    for (String tag : tagNames) {
      NodeList tagNodes = element.getElementsByTagName(tag);
      if (tagNodes.getLength() != 1)
        throw new InvalidModelException("Did not have one and only one "+tag+" element");
      Element tagE = (Element)tagNodes.item(0);
      double val = Double.parseDouble(tagE.getTextContent());
      Units units = Units.getNone();
      if (tagE.hasAttribute("units"))
        units = Units.getUnit(tagE.getAttribute("units"));
      
      body.addExtraParameter(new RobotAttributeValues(tag, new Measurement(units, val)));
    }
    
    return body;
  }
  
  @Override
  public Element toXMLElement(Document doc) {
    Element bodyElement = doc.createElement("body");
    
    // be sure to preserve the library name
    if (!m_libraryName.isEmpty())
      bodyElement.setAttribute("library", m_libraryName);
    
    Element idElement = doc.createElement("id");
    idElement.setTextContent(getId());
    
    Element positionElement = m_position.toXMLElement(doc);
    
    Element rotationElement = null;
    if (m_rotation != null)
      rotationElement = m_rotation.toXMLElement(doc);

    // FIXME mass?
    Element massElement = doc.createElement("mass");
    massElement.setTextContent(Double.toString(m_mass.getStandardValue()));
    
    bodyElement.appendChild(idElement);
    bodyElement.appendChild(positionElement);
    if (rotationElement != null)
      bodyElement.appendChild(rotationElement);
    bodyElement.appendChild(massElement);
    
    // need to add the extra stuff
    for (RobotAttributeValues rav : m_extraParameters.values()) {
      String attribute = rav.getAttribute();
      Units units = rav.getMeasurement().getUnits();
      double value = rav.getMeasurement().getStandardValue();
      Element extraElement = doc.createElement(attribute);
      extraElement.setTextContent(Double.toString(value));
      extraElement.setAttribute("units", units.getBaseUnit().getName());
      bodyElement.appendChild(extraElement);
    }
    
    return bodyElement;
  }
  
  /**
   * @return true iff the {@link ModelBody} has a {@link ComponentLibrary}
   */
  public boolean hasComponentLibrary() {
    return (m_library != null);
  }
  
  /**
   * @return the {@link ComponentLibrary}
   */
  public ComponentLibrary getComponentLibrary() {
    return m_library;
  }
  
  /**
   * @return true iff linked to a particular component in a component library
   */
  public boolean isComponentFixed() {
    return hasComponentInstance();
  }
  
  public boolean hasComponentInstance() {
    return (hasComponentLibrary()  && (m_component != null));
  }

  /**
   * @return the {@link ComponentInstance}
   */
  public ComponentInstance getComponentInstance() {
    return m_component;
  }
  
  public void setComponentLibrary(ComponentLibrary library) {
    m_library = library;
  }
  
  public void setComponentInstance(ComponentInstance component) {
    m_component = component;
  }
}
