/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */

package com.bbn.morph.model;

import java.util.List;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * A box in our model
 * @author aadler
 *
 */
public class ModelBox extends ModelBody {
	
  Measurement m_length;
  Measurement m_width;
  Measurement m_height;
  protected final static String TYPE_NAME = "box";
  
  /**
   * Create a box in our model 
   * @param id the id 
   * @param libraryName the library name for this component or an empty string if none
   * @param position the position
   * @param rotation the rotation, can be null
   * @param box the box (dimensions specified)
   * @param mass the mass of the box
   */
  public ModelBox(String id, String libraryName, ModelPosition position, ModelRotation rotation, Measurement length, 
		  Measurement width, Measurement height, Measurement mass) {
    super(id, libraryName, position, rotation, mass);
    m_length = length;
    m_width = width;
    m_height = height;
    
    // FIXME not sure if I want to update the box to be in the right position or if I just care that it has the right dimensions?
    // FIXME I think the ODE system has to be initialized somewhere
  }
  
  @Override
  public ModelObject copy() {
    ModelRotation rotation = (getRotation() == null)?null:getRotation().copy();
    ModelBox box = new ModelBox(getId(), getLibraryName(), getPosition().copy(), rotation,
    		m_length = new Measurement(Units.getMeters(), m_length.getStandardValue()), 
    		m_width = new Measurement(Units.getMeters(), m_width.getStandardValue()), 
    		m_height = new Measurement(Units.getMeters(), m_height.getStandardValue()), 
    		getMass());
    box.addExtraParameters(getExtraParameters());
    return box;
  }
  
  /* (non-Javadoc)
   * @see com.bbn.morph.model.ModelBody#clone()
   */
  public ModelBox clone() {
    ModelRotation rotation = (getRotation() == null)?null:getRotation().copy();
    ModelBox box = new ModelBox(getId(), getLibraryName(), getPosition().copy(), rotation,
    		m_length = new Measurement(Units.getMeters(), m_length.getStandardValue()), 
    		m_width = new Measurement(Units.getMeters(), m_width.getStandardValue()), 
    		m_height = new Measurement(Units.getMeters(), m_height.getStandardValue()),
    		getMass());
    box.setParentObject(getParentObject());
    box.addExtraParameters(getExtraParameters());
    return box;
  }
  

  @Override
  public Set<RobotAttributeValues> getAttributeValues() {
    Set<RobotAttributeValues> rav = super.getAttributeValues();
    rav.add(new RobotAttributeValues(getId()+".lengthx",m_length));
    rav.add(new RobotAttributeValues(getId()+".lengthy",m_width));
    rav.add(new RobotAttributeValues(getId()+".lengthz",m_height));
    return rav;
  }
  

  
  @Override
  public Measurement getAttributeValue(String attribute) {
    if (attribute.equals("lengthx"))
      return m_length;
    else if (attribute.equals("lengthy"))
      return m_width;
    else if (attribute.equals("lengthz"))
      return m_height;
    else
      // try our super class:
      return super.getAttributeValue(attribute);
  }
  
  @Override
  public void updateRobotAttributeFromDeltas(RobotAttributeValues robotAttributeValues,
      List<ModelJoint> joints) {
    // handle what we can handle:
    String attribute = robotAttributeValues.getAttribute();
    double delta = robotAttributeValues.getMeasurement().getStandardValue();
    ModelPosition jointsDirection = getJointsDirection(joints);
    
    if (attribute.equals("lengthx")) {
      if (jointsDirection.getxPos().getStandardValue() != 0) 
        // shift the x position by half the delta away from the joint
        updateRobotAttributeFromDeltas((new RobotAttributeValues("x", jointsDirection.getxPos().multMD(-1*delta/2))), joints);
      m_length = m_length.addMD(delta);
    }
    else if (attribute.equals("lengthy")) {
      if (jointsDirection.getyPos().getStandardValue() != 0) 
        // shift the y position by half the delta away from the joint
        updateRobotAttributeFromDeltas((new RobotAttributeValues("y", jointsDirection.getyPos().multMD(-1*delta/2))), joints);
      m_width = m_width.addMD(delta);
    }
    else if (attribute.equals("lengthz")) {
      // new z length:
      Measurement newZlength = m_height.addMD(delta);

      if (jointsDirection.getzPos().getStandardValue() != 0) 
        // shift the z position by half the delta away from the joint
        updateRobotAttributeFromDeltas((new RobotAttributeValues("z", jointsDirection.getzPos().multMD(-1*delta/2))), joints);
      else {
        // TODO we need to figure a better way to do this I think see Ticket #29
        Measurement deltaZ = (newZlength.divideMD(2.0)).subtract(getAttributeValue("z"));

        if (deltaZ.getStandardValue() > 0)
          // this is to keep the box above or on the z=0 plane
          updateRobotAttributeFromDeltas((new RobotAttributeValues("z", deltaZ)), joints);

      }
      m_height = newZlength;
    }
    else
      // try our super class:
      super.updateRobotAttributeFromDeltas(robotAttributeValues, joints);
  }

  /**
   * Creates a {@link ModelBox} from the given XML element
   * @param element the XML element representing a {@link ModelBox}
   * @param id the id
   * @param libraryName the library name for this component or an empty string if none
   * @param position the position of the box
   * @param rotation the rotation of the box, can be null
   * @param mass the mass
   * @param tagNames the set of tag names that have not been handled (tags removed as they are handled)
   * @return the {@link ModelBox} encoded by the XML
   * @throws InvalidModelException if the XML is ill-formed
   */
  public static ModelBody createFromXML(Element element, String id, String libraryName,
      ModelPosition position, ModelRotation rotation, Measurement mass, Set<String> tagNames) throws InvalidModelException {
    NodeList dimNodes = element.getElementsByTagName("dim");
    if (dimNodes.getLength() != 1)
      throw new InvalidModelException("Did not have one and only one position dim");
    Element dim = ((Element)dimNodes.item(0));
    
    // FIXME -- these should really be x y z not w h b
    if (!dim.hasAttribute("w") || !dim.hasAttribute("h") ||
        !dim.hasAttribute("b"))
      throw new InvalidModelException("Missing required position attribute");
    
    try {
      double x = Double.parseDouble(dim.getAttribute("w"));
      double y = Double.parseDouble(dim.getAttribute("b"));
      double z = Double.parseDouble(dim.getAttribute("h"));
      
      Measurement l = new Measurement(Units.getMeters(), x);
      Measurement w = new Measurement(Units.getMeters(), y);
      Measurement h = new Measurement(Units.getMeters(), z);
      
      tagNames.remove("dim");
      
      return new ModelBox(id, libraryName, position, rotation, l, w, h, mass);
      
    }
    catch (NumberFormatException e) {
      throw new InvalidModelException(e);
    }
  }
  
  @Override
  public Element toXMLElement(Document doc) {
    Element baseElement = super.toXMLElement(doc);
    baseElement.setAttribute("type", TYPE_NAME);
    
    Element dim = doc.createElement("dim");
    dim.setAttribute("w", Double.toString(m_length.getStandardValue()));
    dim.setAttribute("b", Double.toString(m_width.getStandardValue()));
    dim.setAttribute("h", Double.toString(m_height.getStandardValue()));
    
    baseElement.appendChild(dim);
    return baseElement;
  }

}
