/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.model;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Represents a hinge joint in our model
 * @author aadler
 *
 */
public class ModelHingeJoint extends ModelJoint {

  // FIXME I think this name should be hingeJoint to make the others!
  protected static final String TYPE_NAME = "hinged";
  
  private Double m_axisVx;
  private Double m_axisVy;
  private Double m_axisVz;
  private Double m_anchorVx;
  private Double m_anchorVy;
  private Double m_anchorVz;
  private Double m_lowStop;
  private Double m_highStop;
  
  /**
   * Creates a hinge joint
   * @param bodyId1 the id of the first body
   * @param bodyId2 the id of the second body
   * @param anchorV the anchor position
   * @param axisV the axis position
   * @param lowStop the low stop
   * @param highStop the high stop
   */
  public ModelHingeJoint(String bodyId1, String bodyId2, Double anchorVx, 
      Double anchorVy, Double anchorVz, Double axisVx, Double axisVy, 
      Double axisVz, Double lowStop, Double highStop) {
    super(bodyId1, bodyId2);
    m_anchorVx = anchorVx;
    m_anchorVy = anchorVy;
    m_anchorVz = anchorVz;
    m_axisVx = axisVx;
    m_axisVy = axisVy;
    m_axisVz = axisVz;
    m_lowStop = lowStop;
    m_highStop = highStop;
  }
  
  @Override
  public ModelObject copy() {
    return new ModelHingeJoint(getBodyId1(), getBodyId2(), getAnchorVx(), 
        getAnchorVy(), getAnchorVz(), getAxisVx(), getAxisVy(), getAxisVz(), 
        getLowStop(), getHighStop());
  }
  
  public Double getAnchorVx() {
    return m_anchorVx;
  }
  
  public Double getAnchorVy() {
    return m_anchorVy;
  }
  
  public Double getAnchorVz() {
    return m_anchorVz;
  }
  
  public Double getAxisVx() {
    return m_axisVx;
  }
  
  public Double getAxisVy() {
    return m_axisVy;
  }

  public Double getAxisVz() {
    return m_axisVz;
  }

  public Double getHighStop() {
    return m_highStop;
  }
  
  public Double getLowStop() {
    return m_lowStop;
  }
  
  /**
   * Creates a {@link ModelHingeJoint} from the given XML element
   * @param element the XML element representing a {@link ModelHingeJoint}
   * @param id1 the first id
   * @param id1 the second id
   * @return the {@link ModelHingeJoint} encoded by the XML
   * @throws InvalidModelException if the XML is ill-formed
   */
  public static ModelJoint createFromXML(Element element, String id1, 
      String id2) throws InvalidModelException {
    NodeList anchorNodes = element.getElementsByTagName("anchor");
    if (anchorNodes.getLength() != 1)
      throw new InvalidModelException("Did not have one and only one anchor");
    Element anchor = ((Element)anchorNodes.item(0));
    
    NodeList axisNodes = element.getElementsByTagName("axis");
    if (axisNodes.getLength() != 1)
      throw new InvalidModelException("Did not have one and only one axis");
    Element axis = ((Element)axisNodes.item(0));
    
    // FIXME are these required???
    
    NodeList lowNodes = element.getElementsByTagName("lowStop");
    Element lowNode = null;
    if (lowNodes.getLength() == 1)
      lowNode = ((Element)lowNodes.item(0));

    NodeList highNodes = element.getElementsByTagName("highStop");
    Element highNode = null;
    if (highNodes.getLength() == 1)
      highNode = ((Element)highNodes.item(0));

    try {
      if (!anchor.hasAttribute("x") || !anchor.hasAttribute("y") ||
          !anchor.hasAttribute("z"))
        throw new InvalidModelException("Missing required anchor attribute");
      double anchorX = Double.parseDouble(anchor.getAttribute("x"));
      double anchorY = Double.parseDouble(anchor.getAttribute("y"));
      double anchorZ = Double.parseDouble(anchor.getAttribute("z"));
    
      if (!axis.hasAttribute("x") || !axis.hasAttribute("y") ||
          !axis.hasAttribute("z"))
        throw new InvalidModelException("Missing required anchor attribute");
      double axisX = Double.parseDouble(axis.getAttribute("x"));
      double axisY = Double.parseDouble(axis.getAttribute("y"));
      double axisZ = Double.parseDouble(axis.getAttribute("z"));
      
      Double lowStop = (lowNode == null)?null:Double.parseDouble(lowNode.getTextContent());
      Double highStop = (highNode == null)?null:Double.parseDouble(highNode.getTextContent());
    
      return new ModelHingeJoint(id1, id2, anchorX, anchorY, anchorZ, axisX, axisY, axisZ, lowStop, highStop);
    }
    catch (NumberFormatException e) {
      throw new InvalidModelException(e);
    }
  }

  @Override
  public Element toXMLElement(Document doc) {
    Element baseElement = super.toXMLElement(doc);
    baseElement.setAttribute("type", TYPE_NAME);
    
    Element anchor = doc.createElement("anchor");
    anchor.setAttribute("x", Double.toString(getAnchorVx()));
    anchor.setAttribute("y", Double.toString(getAnchorVy()));
    anchor.setAttribute("z", Double.toString(getAnchorVz()));
    baseElement.appendChild(anchor);
    
    Element axis = doc.createElement("axis");
    axis.setAttribute("x", Double.toString(getAxisVx()));
    axis.setAttribute("y", Double.toString(getAxisVy()));
    axis.setAttribute("z", Double.toString(getAxisVz()));
    baseElement.appendChild(axis);

    if (m_lowStop != null) {
      Element lowStop = doc.createElement("lowStop");
      lowStop.setTextContent(Double.toString(m_lowStop));
      baseElement.appendChild(lowStop);
    }
    
    if (m_highStop != null) {
      Element highStop = doc.createElement("highStop");
      highStop.setTextContent(Double.toString(m_highStop));
      baseElement.appendChild(highStop);
    }
    
    return baseElement;
  }
}
