/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.model;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * An abstract class that represents a joint. Subclasses are:
 * {@link ModelHingeJoint}, {@link ModelFixedJoint}, and {@link ModelSliderJoint}.
 * @author aadler
 *
 */
public abstract class ModelJoint extends ModelObject {

  private String m_bodyId1;
  private String m_bodyId2;
  
  /**
   * Creates a joint between the two specified bodies
   * @param bodyId1 the id of the first body
   * @param bodyId2 the id of the second body
   */
  public ModelJoint(String bodyId1, String bodyId2) {
    m_bodyId1 = bodyId1;
    m_bodyId2 = bodyId2;
  }
  
  /**
   * @return the id of the first body
   */
  public String getBodyId1() {
    return m_bodyId1;
  }
  
  /**
   * @return the id of the second body
   */
  public String getBodyId2() {
    return m_bodyId2;
  }
  
  /**
   * Creates a {@link ModelJoint} from the given XML element
   * @param element the XML element representing a {@link ModelJoint}
   * @return the {@link ModelJoint} encoded by the XML
   * @throws InvalidModelException if the XML is ill-formed
   */
  public static ModelJoint createFromXML(Element element) throws InvalidModelException {
    if (!element.hasAttribute("type"))
      throw new InvalidModelException("joint element missing type attribute");
    
    String type = element.getAttribute("type");
    
    // FIXME in schema this would just be two elements of bodyId, but in file it is "a" and "b" -- we need to fix this
    NodeList aNodes = element.getElementsByTagName("a");
    if (aNodes.getLength() != 1)
      throw new InvalidModelException("Did not have one and only one a element");
    String id1 = ((Element)aNodes.item(0)).getTextContent();
    NodeList bNodes = element.getElementsByTagName("b");
    if (bNodes.getLength() != 1)
      throw new InvalidModelException("Did not have one and only one b element");
    String id2 = ((Element)bNodes.item(0)).getTextContent();
    
    if (type.equals(ModelFixedJoint.TYPE_NAME)) {
      return ModelFixedJoint.createFromXML(element, id1, id2);
    }
    else if (type.equals(ModelHingeJoint.TYPE_NAME)) {
      return ModelHingeJoint.createFromXML(element, id1, id2);
    }
    else if (type.equals(ModelSliderJoint.TYPE_NAME)) {
      return ModelSliderJoint.createFromXML(element, id1, id2);
    }
    else 
      throw new RuntimeException("Unhandled type: "+type);
  }
  
  @Override
  public Element toXMLElement(Document doc) {
    Element bodyElement = doc.createElement("joint");
    
    // FIXME a and b or "bodyId"
    
    Element id1Element = doc.createElement("a");
    id1Element.setTextContent(getBodyId1());
    
    Element id2Element = doc.createElement("b");
    id2Element.setTextContent(getBodyId2());
    
    bodyElement.appendChild(id1Element);
    bodyElement.appendChild(id2Element);
    
    return bodyElement;
  }

}
