/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */

package com.bbn.morph.model;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Representing a position in our model
 * @author aadler
 *
 */
public class ModelPosition {
  
  private Measurement m_xPos;
  private Measurement m_yPos;
  private Measurement m_zPos;

  /**
   * Creates a position at the specified location
   * @param x x position
   * @param y y position
   * @param z z position
   */
  public ModelPosition(Measurement x, Measurement y, Measurement z) {
    m_xPos = x;
    m_yPos = y;
    m_zPos = z;
  }
  
  /**
   * @return a new {@link ModelPosition} that is a copy of this one
   */
  public ModelPosition copy() {
    return new ModelPosition(m_xPos, m_yPos, m_zPos);
  }
  
  /**
   * @return the x position
   */
  public Measurement getxPos() {
    return m_xPos;
  }
  
  /**
   * @return the y position
   */
  public Measurement getyPos() {
    return m_yPos;
  }
  
  /**
   * @return the z position
   */
  public Measurement getzPos() {
    return m_zPos;
  }
  
  /**
   * @param xPos set the x position
   */
  public void setxPos(Measurement xPos) {
    m_xPos = xPos;
  }
  
  /**
   * @param yPos set the y position
   */
  public void setyPos(Measurement yPos) {
    m_yPos = yPos;
  }
  
  /**
   * @param zPos set the z position
   */
  public void setzPos(Measurement zPos) {
    m_zPos = zPos;
  }

  /**
   * @param doc the XML document
   * @return the {@link ModelPosition} as an XML {@link Element} in the {@link Document}
   */
  public Element toXMLElement(Document doc) {
    Element positionElement = doc.createElement("position");
    positionElement.setAttribute("x", Double.toString(getxPos().getStandardValue()));
    positionElement.setAttribute("y", Double.toString(getyPos().getStandardValue()));
    positionElement.setAttribute("z", Double.toString(getzPos().getStandardValue()));
    return positionElement;
  }
  

/**
   * Creates a {@link ModelPosition} from the given XML element
   * @param positionNode the XML element representing a {@link ModelPosition}
   * @return the {@link ModelPosition} encoded by the XML
   * @throws InvalidModelException if the XML is ill-formed
   */
  public static ModelPosition createFromXML(Element positionNode) throws InvalidModelException {
    if (!positionNode.hasAttribute("x") || !positionNode.hasAttribute("y") ||
        !positionNode.hasAttribute("z"))
      throw new InvalidModelException("Missing required position attribute");
    
    try {
      double x = Double.parseDouble(positionNode.getAttribute("x"));
      double y = Double.parseDouble(positionNode.getAttribute("y"));
      double z = Double.parseDouble(positionNode.getAttribute("z"));
      
      Measurement xMeters = new Measurement(Units.getMeters(), x);
      Measurement yMeters = new Measurement(Units.getMeters(), y);
      Measurement zMeters = new Measurement(Units.getMeters(), z);
       
      
      return new ModelPosition(xMeters, yMeters, zMeters);
    }
    catch (NumberFormatException e) {
      throw new InvalidModelException(e);
    }
  }
  
  
}
