/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.model;

import java.util.ArrayList;
import java.util.Collection;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class ModelQuaternion extends ModelRotation {

  public static final String TYPE_NAME = "quaternion";
  
  private Measurement m_q1;
  private Measurement m_q2;
  private Measurement m_q3;
  private Measurement m_q4;
  
  public ModelQuaternion(double q1, double q2, double q3, double q4) {
    m_q1 = new Measurement(Units.getRadians(), q1);
    m_q2 = new Measurement(Units.getRadians(), q2);
    m_q3 = new Measurement(Units.getRadians(), q3);
    m_q4 = new Measurement(Units.getRadians(), q4);
  }
  
  @Override
  public Collection<RobotAttributeValues> getAttributes(String prefix) {
    ArrayList<RobotAttributeValues> attList = new ArrayList<RobotAttributeValues>();
    attList.add(new RobotAttributeValues(prefix+".q1", m_q1));
    attList.add(new RobotAttributeValues(prefix+".q2", m_q2));
    attList.add(new RobotAttributeValues(prefix+".q3", m_q3));
    attList.add(new RobotAttributeValues(prefix+".q4", m_q4));
    return attList;
  }
  
  @Override
  public ModelRotation copy() {
    return new ModelQuaternion(m_q1.getStandardValue(), m_q2.getStandardValue(), m_q3.getStandardValue(), m_q4.getStandardValue());
  }
  
  @Override
  public boolean isHandledAttribute(String attribute) {
    if (attribute.equals("q1") || attribute.equals("q2") ||
        attribute.equals("q3") || attribute.equals("q4"))
      return true;
    else
      return false;
  }
  
  @Override
  public Measurement getAttributeValue(String attribute) {
    if (attribute.equals("q1"))
      return m_q1;
    else if (attribute.equals("q2"))
      return m_q1;
    else if (attribute.equals("q3"))
      return m_q3;
    else if (attribute.equals("q4"))
      return m_q4;
    return null;
  }
  
  @Override
  public void updateAttributeByDelta(String attribute, double delta) {
    if (attribute.equals("q1"))
      m_q1 = m_q1.addMD(delta);
    else if (attribute.equals("q2"))
      m_q2 = m_q2.addMD(delta);
    else if (attribute.equals("q3"))
      m_q3 = m_q3.addMD(delta);
    else if (attribute.equals("q4"))
      m_q4 = m_q4.addMD(delta);
  }
  
  /* (non-Javadoc)
   * @see com.bbn.morph.model.ModelRotation#toXMLElement(org.w3c.dom.Document)
   */
  public Element toXMLElement(Document doc) {
    Element positionElement = doc.createElement(TYPE_NAME);
    positionElement.setAttribute("q1", Double.toString(m_q1.getStandardValue()));
    positionElement.setAttribute("q2", Double.toString(m_q2.getStandardValue()));
    positionElement.setAttribute("q3", Double.toString(m_q3.getStandardValue()));
    positionElement.setAttribute("q4", Double.toString(m_q4.getStandardValue()));
    return positionElement;
  }
  
  
  /**
   * Creates a {@link ModelQuaternion} from the given XML element
   * @param rotationNode the XML element representing a {@link ModelQuaternion}
   * @return the {@link ModelQuaternion} encoded by the XML
   * @throws InvalidModelException if the XML is ill-formed
   */
  public static ModelQuaternion createFromXML(Element rotationNode) throws InvalidModelException {
    if (!rotationNode.hasAttribute("q1") || !rotationNode.hasAttribute("q2") ||
        !rotationNode.hasAttribute("q3") || !rotationNode.hasAttribute("q4"))
      throw new InvalidModelException("Missing required position attribute");
    
    try {
      double q1 = Double.parseDouble(rotationNode.getAttribute("q1"));
      double q2 = Double.parseDouble(rotationNode.getAttribute("q2"));
      double q3 = Double.parseDouble(rotationNode.getAttribute("q3"));
      double q4 = Double.parseDouble(rotationNode.getAttribute("q4"));
      
      return new ModelQuaternion(q1, q2, q3, q4);
    }
    catch (NumberFormatException e) {
      throw new InvalidModelException(e);
    }
  }
}
