/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.model;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class ModelUtils {

  /**
   * Write the robot model to the specified file as XML
   * @param file the file to write the XML to
   * @param objects the objects to write to the file
   * @throws Exception if an error writing to the file occurs
   */
  public static void writeToXMLFile(File file, List<ModelObject> objects) throws Exception {
    DocumentBuilderFactory documentFactory = DocumentBuilderFactory.newInstance();
    DocumentBuilder documentBuilder = documentFactory.newDocumentBuilder();
  
    //root elements
    Document doc = documentBuilder.newDocument();
    Element rootElement = doc.createElement("root");
    
    for (ModelObject modelObject : objects) {
      rootElement.appendChild(modelObject.toXMLElement(doc));
    }
    doc.appendChild(rootElement);
    
    // set up a transformer:
    Transformer transformer = TransformerFactory.newInstance().newTransformer();
    transformer.setOutputProperty(OutputKeys.INDENT, "yes");
    transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
    
    // write out to a file
    DOMSource source = new DOMSource(doc);
    StreamResult result = new StreamResult(file);
    transformer.transform(source, result);
  }

  private static Map<String, Double> constructMap(List<ModelObject> allObjects) {
    Map<String, Double> map = new HashMap<String, Double>();
    
    for (ModelObject obj : allObjects) {
      // TODO we are just dealing with bodies for now:
      if (obj instanceof ModelBody) {
        ModelBody b = (ModelBody)obj;
        //String id = b.getId();
        for (RobotAttributeValues rav : b.getAttributeValues()) {
          String key = rav.getAttribute();
          double standardValue = rav.getMeasurement().getStandardValue();
          map.put(key, standardValue);
        }
      }
    }
    return map;
  }
  
  private static Map<String, ModelBody> constructObjectMap(List<ModelObject> allObjects) {
    Map<String, ModelBody> map = new HashMap<String, ModelBody>();
    
    for (ModelObject obj : allObjects) {
      // TODO we are just dealing with bodies for now:
      if (obj instanceof ModelBody) {
        ModelBody b = (ModelBody)obj;
        String id = b.getId();
        map.put(id, b);
      }
    }
    return map;
  }
  
  private static void writeDimensions(BufferedWriter bw, Map<String, Double> valueMap) throws IOException {
    bw.write("  <xacro:property name=\"body_width\" value=\""+valueMap.get("botbody.lengthy")+"\" />"); bw.newLine();
    bw.write("  <xacro:property name=\"body_length\" value=\""+valueMap.get("botbody.lengthx")+"\" />"); bw.newLine();
    bw.write("  <xacro:property name=\"body_height\" value=\""+valueMap.get("botbody.lengthz")+"\" />"); bw.newLine();
    bw.newLine();
    // Just use one wheel to pull these in:
    bw.write("  <xacro:property name=\"wheel_radius\" value=\""+valueMap.get("leftPassiveWheel.radius")+"\" />"); bw.newLine();
    bw.write("  <xacro:property name=\"wheel_length\" value=\""+valueMap.get("leftPassiveWheel.length")+"\" />"); bw.newLine();
    bw.newLine();
    // FIXME this only deals with one arm!
    bw.write("  <xacro:property name=\"flipper_width\" value=\""+valueMap.get("leftarm.lengthy")+"\" />"); bw.newLine();
    bw.write("  <xacro:property name=\"flipper_length\" value=\""+valueMap.get("leftarm.lengthz")+"\" />"); bw.newLine();
    bw.write("  <xacro:property name=\"flipper_height\" value=\""+valueMap.get("leftarm.lengthx")+"\"/>"); bw.newLine();
    bw.newLine();
    bw.write("  <xacro:property name=\"wheel_buffer\" value=\"0.0025\" />"); bw.newLine();
    bw.write("  <xacro:property name=\"drive_wheel_offset\" value=\"0.021082\" />"); bw.newLine();
    bw.write("  <xacro:property name=\"idle_wheel_offset\" value=\"0.019050\" />"); bw.newLine();

  }
  
  private static void writeWheels(BufferedWriter bw, Map<String, ModelBody> objMap, Map<String, Double> valueMap) throws IOException {
    int i = 0;
    while (objMap.containsKey("leftWheel"+i)) {
      ModelBody wheel = objMap.get("leftWheel"+i);
      ModelObject xForm = wheel.getParentObject();
      if (xForm != null) {
        double shift = -((ModelTransform)xForm).getXShift();
        bw.write("  <xacro:wheel prefix=\"left_drive"+i+"\" reflect=\"1\" position=\""+shift+"\" />"); bw.newLine();
        bw.write("  <xacro:wheel prefix=\"right_drive"+i+"\" reflect=\"1\" position=\""+shift+"\" />"); bw.newLine();
        //System.out.println("leftWheel"+i+" position is "+shift);
      }
      
      i++;
    }
    
    // passive wheel:
    ModelBody wheel = objMap.get("leftPassiveWheel");
    ModelObject xForm = wheel.getParentObject();
    if (xForm != null) {
      double shift = ((ModelTransform)xForm).getXShift();
      double bodyX = valueMap.get("botbody.x");
      double idleShift = bodyX - shift;
      bw.write("  <xacro:wheel prefix=\"left_idle\" reflect=\"-1\" position=\""+idleShift+"\" />"); bw.newLine();
      bw.write("  <xacro:wheel prefix=\"right_idle\" reflect=\"1\" position=\""+idleShift+"\" />"); bw.newLine();
      //System.out.println("leftPassiveWheel posistion is "+(shift - bodyX));
      //System.out.println();
      //System.out.println("leftPassiveWheel posistion is "+valueMap.get("leftPassiveWheel.x"));
    }
  }
  
  /**
   * Write the robot model to the specified file as ROS XACRO
   * @param file the file to write the XACRO to
   * @param allObjects all the objects in the models to write to the file
   * @throws Exception if an error writing to the file occurs
   */
  public static void writeToROSFile(File file, List<ModelObject> allObjects) throws Exception {
    // FIXME this is really, really hacky at the moment. See bug #???
    
    Map<String, Double> valueMap = constructMap(allObjects);
    Map<String, ModelBody> bodyMap = constructObjectMap(allObjects);
    
    FileWriter fw = new FileWriter(file);
    BufferedWriter bw = new BufferedWriter(fw);
    
    bw.write("<?xml version=\"1.0\"?>"); bw.newLine();
    bw.write("<robot "); bw.newLine();
    bw.write("       xmlns:xi=\"http://www.w3.org/2001/XInclude\""); bw.newLine();
    bw.write("    xmlns:gazebo=\"http://playerstage.sourceforge.net/gazebo/xmlschema/#gz\""); bw.newLine();
    bw.write("    xmlns:model=\"http://playerstage.sourceforge.net/gazebo/xmlschema/#model\""); bw.newLine();
    bw.write("    xmlns:sensor=\"http://playerstage.sourceforge.net/gazebo/xmlschema/#sensor\""); bw.newLine();
    bw.write("    xmlns:body=\"http://playerstage.sourceforge.net/gazebo/xmlschema/#body\""); bw.newLine();
    bw.write("    xmlns:geom=\"http://playerstage.sourceforge.net/gazebo/xmlschema/#geom\""); bw.newLine();
    bw.write("    xmlns:joint=\"http://playerstage.sourceforge.net/gazebo/xmlschema/#joint\""); bw.newLine();
    bw.write("    xmlns:interface=\"http://playerstage.sourceforge.net/gazebo/xmlschema/#interface\""); bw.newLine();
    bw.write("    xmlns:rendering=\"http://playerstage.sourceforge.net/gazebo/xmlschema/#rendering\""); bw.newLine();
    bw.write("    xmlns:renderable=\"http://playerstage.sourceforge.net/gazebo/xmlschema/#renderable\""); bw.newLine();
    bw.write("    xmlns:controller=\"http://playerstage.sourceforge.net/gazebo/xmlschema/#controller\""); bw.newLine();
    bw.write("    xmlns:physics=\"http://playerstage.sourceforge.net/gazebo/xmlschema/#physics\""); bw.newLine();
    bw.write("    xmlns:xacro=\"http://playerstage.sourceforge.net/gazebo/xmlschema/#interface\" name=\"minidroid\">"); bw.newLine();
    
    bw.newLine();
    bw.write("  <!-- Parameters for robot dimensions -->"); bw.newLine();
    
    writeDimensions(bw, valueMap);
    
    bw.newLine();
    bw.write("  <xacro:property name=\"idle_axle\" value=\"${(-body_length/2) + idle_wheel_offset}\" />"); bw.newLine();
    bw.write("  <xacro:property name=\"drive_axle\" value=\"${body_length/2 - drive_wheel_offset}\" />"); bw.newLine();
    bw.newLine();
    bw.write("  <xacro:property name=\"pi\" value=\"3.1415\" />"); bw.newLine();
    bw.newLine();
    bw.write("  <!-- Macro for assigning identity matrix as moment of intertia -->"); bw.newLine();
    bw.write("  <xacro:macro name=\"default_inertial\" params=\"mass\">"); bw.newLine();
    bw.write("    <inertial>"); bw.newLine();
    bw.write("      <mass value=\"${mass}\" />"); bw.newLine();
    bw.write("      <inertia ixx=\"1.0\" ixy=\"0.0\" ixz=\"0.0\""); bw.newLine();
    bw.write("               iyy=\"1.0\" iyz=\"0.0\""); bw.newLine();
    bw.write("               izz=\"1.0\" />"); bw.newLine();
    bw.write("    </inertial>"); bw.newLine();
    bw.write("  </xacro:macro>"); bw.newLine();
    bw.newLine();
    bw.newLine();
    bw.newLine();
    bw.newLine();
    bw.write("  <!-- Body -->"); bw.newLine();
    bw.write("  <link name=\"base_link\">"); bw.newLine();
    bw.write("    <visual>"); bw.newLine();
    bw.write("      <geometry>"); bw.newLine();
    bw.write("        <box size=\"${body_width} ${body_length} ${body_height}\"/>"); bw.newLine();
    bw.write("      </geometry>"); bw.newLine();
    bw.write("      <material name=\"blue\">"); bw.newLine();
    bw.write("        <color rgba=\"0 0 .8 1\"/>"); bw.newLine();
    bw.write("      </material>"); bw.newLine();
    bw.write("    </visual>"); bw.newLine();
    bw.write("    <collision>"); bw.newLine();
    bw.write("      <geometry>"); bw.newLine();
    bw.write("        <box size=\"${body_width} ${body_length} ${body_height}\"/>"); bw.newLine();
    bw.write("      </geometry>"); bw.newLine();
    bw.write("    </collision>"); bw.newLine();
    bw.write("    <xacro:default_inertial mass=\"10\"/>"); bw.newLine();
    bw.write("  </link>"); bw.newLine();
    bw.newLine();
    bw.newLine();
    bw.write("  <!-- Macro for placing wheels on robot, includes wheel and joint -->"); bw.newLine();
    bw.write("  <xacro:macro name=\"wheel\" params=\"prefix reflect position\">"); bw.newLine();
    bw.write("    <link name=\"${prefix}_wheel\">"); bw.newLine();
    bw.write("      <visual>"); bw.newLine();
    bw.write("    <geometry>"); bw.newLine();
    bw.write("      <cylinder radius=\"${wheel_radius}\" length=\"${wheel_length}\" />"); bw.newLine();
    bw.write("  </geometry>"); bw.newLine();
    bw.write("  <origin xyz=\"0 0 0\" rpy=\"0 ${pi/2} 0\"/>"); bw.newLine();
    bw.write("  <material name=\"white\">"); bw.newLine();
    bw.write("    <color rgba=\"1 1 1 1\"/>"); bw.newLine();
    bw.write("  </material>"); bw.newLine();
    bw.write("      </visual>"); bw.newLine();
    bw.write("      <collision>"); bw.newLine();
    bw.write("  <geometry>"); bw.newLine();
    bw.write("    <cylinder radius=\"${wheel_radius}\" length=\"${wheel_length}\" />"); bw.newLine();
    bw.write("  </geometry>  "); bw.newLine();
    bw.write("  <origin xyz=\"0 0 0\" rpy=\"0 ${pi/2} 0\"/>"); bw.newLine();
    bw.write("      </collision>"); bw.newLine();
    bw.write("      <xacro:default_inertial mass=\"1\"/>"); bw.newLine();
    bw.write("    </link>"); bw.newLine();
    bw.newLine();
    bw.write("    <joint name=\"base_to_${prefix}_wheel\" type=\"continuous\">"); bw.newLine();
    bw.write("      <axis xyz=\"1 0 0\" />"); bw.newLine();
    bw.write("      <parent link=\"base_link\"/>"); bw.newLine();
    bw.write("      <child link=\"${prefix}_wheel\"/>"); bw.newLine();
    bw.write("      <origin xyz=\"${reflect*(wheel_length+body_width+wheel_buffer)/2} ${position} 0\" />"); bw.newLine();
    bw.write("    </joint>"); bw.newLine();
    bw.write("  </xacro:macro>"); bw.newLine();
    bw.newLine();
    bw.write("    <!-- Macro for placing flippers on robot, includes joints -->"); bw.newLine();
    bw.write("    <xacro:macro name=\"flipper\" params=\"prefix reflect\">"); bw.newLine();
    bw.write("      <link name=\"${prefix}_flipper\">"); bw.newLine();
    bw.write("        <visual>"); bw.newLine();
    bw.write("    <geometry>"); bw.newLine();
    bw.write("      <box size=\"${flipper_width} ${flipper_length} ${flipper_height}\" />"); bw.newLine();
    bw.write("    </geometry>"); bw.newLine();
    bw.write("    <origin xyz=\"0 ${flipper_length/2} 0\" rpy=\"0 0 0\"/>"); bw.newLine();
    bw.write("    <material name=\"grey\">"); bw.newLine();
    bw.write("      <color rgba=\"0.5 0.5 0.5 1\"/>"); bw.newLine();
    bw.write("    </material>"); bw.newLine();
    bw.write("        </visual>"); bw.newLine();
    bw.write("        <collision>"); bw.newLine();
    bw.write("    <geometry>"); bw.newLine();
    bw.write("      <box size=\"${flipper_width} ${flipper_length} ${flipper_height}\" />"); bw.newLine();
    bw.write("    </geometry>  "); bw.newLine();
    bw.write("    <origin xyz=\"0  ${flipper_length/2} 0\" rpy=\"0 0 0\"/>"); bw.newLine();
    bw.write("        </collision>"); bw.newLine();
    bw.write("        <xacro:default_inertial mass=\"1\"/>"); bw.newLine();
    bw.write("      </link>"); bw.newLine();
    bw.newLine();
    bw.write("      <joint name=\"base_to_${prefix}_flipper\" type=\"continuous\">"); bw.newLine();
    bw.write("        <axis xyz=\"1 0 0\" />"); bw.newLine();
    bw.write("  <!--   <limit effort=\"1000.0\" lower=\"0.0\" upper=\"0.548\" velocity=\"0.5\"/> -->"); bw.newLine();
    bw.write("        <parent link=\"base_link\"/>"); bw.newLine();
    bw.write("        <child link=\"${prefix}_flipper\"/>"); bw.newLine();
    bw.write("        <origin xyz=\"${reflect*(body_width/2 + wheel_length + wheel_buffer + flipper_width/2)} ${idle_axle} 0\" />"); bw.newLine();
    bw.write("      </joint> "); bw.newLine();
    bw.write("    </xacro:macro>"); bw.newLine();
    bw.newLine();
    bw.newLine();
    bw.write("    <xacro:flipper prefix=\"left\" reflect=\"1\" />"); bw.newLine();
    bw.write("    <xacro:flipper prefix=\"right\" reflect=\"-1\" />"); bw.newLine();
    bw.newLine();
    
    writeWheels(bw, bodyMap, valueMap);

    bw.newLine();
    
    bw.write(" <gazebo reference=\"left_flipper\">"); bw.newLine();
    bw.write("    <material>Gazebo/Red</material>"); bw.newLine();
    bw.write("  </gazebo>"); bw.newLine();
    bw.newLine();
    bw.write("  <gazebo>"); bw.newLine();
    bw.write("      <!-- PR2_ACTARRAY -->"); bw.newLine();
    bw.write("      <controller:gazebo_ros_controller_manager name=\"gazebo_ros_controller_manager\" plugin=\"libgazebo_ros_controller_manager.so\">"); bw.newLine();
    bw.write("        <alwaysOn>true</alwaysOn>"); bw.newLine();
    bw.write("        <updateRate>1000.0</updateRate>"); bw.newLine();
    bw.write("      </controller:gazebo_ros_controller_manager>"); bw.newLine();
    bw.newLine();
    bw.write("      <!-- gazebo_ros_p3d for position groundtruth -->"); bw.newLine();
    bw.write("      <controller:gazebo_ros_p3d name=\"p3d_link1_controller\" plugin=\"libgazebo_ros_p3d.so\">"); bw.newLine();
    bw.write("          <alwaysOn>true</alwaysOn>"); bw.newLine();
    bw.write("          <updateRate>1000.0</updateRate>"); bw.newLine();
    bw.write("          <bodyName>base_link</bodyName>"); bw.newLine();
    bw.write("          <topicName>base_link_pose</topicName>"); bw.newLine();
    bw.write("          <frameName>map</frameName>"); bw.newLine();
    bw.write("          <interface:position name=\"p3d_link1_position\"/>"); bw.newLine();
    bw.write("      </controller:gazebo_ros_p3d>"); bw.newLine();
    bw.write("  </gazebo>"); bw.newLine();
    bw.write("  <!-- mechanism controls -->"); bw.newLine();
    bw.write("  <include filename=\"./madv.transmission.xacro\" />"); bw.newLine();
    
    bw.newLine();
    bw.newLine();
    bw.newLine();
    bw.write("  </robot>");
    

    // Close up
    bw.flush();
    bw.close();
    fw.close();
  }
}
