/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.model;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.EventObject;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.bbn.morph.view.AttributeValueView;
import com.bbn.morph.view.RobotAttribute;

/**
 * A physical model
 * @author aadler
 *
 */
public class PhysicalModel {

  // TODO -- keep track of bodies, joints, transforms, also id -> object map
  // TODO these should be sorted by index number so we can reproduce the XML file
  private Map<String, ModelBody> m_idMap;
  private List<ModelObject> m_topObjects;  
  private List<ModelObject> m_allObjects;
  private List<ModelBody> m_bodies;
  private List<ModelJoint> m_joints;
  
  private List<ModelListener> m_listeners;
  
  /**
   * Creates a {@link PhysicalModel} from the specified {@link ModelObject}s
   * @param modelObjects the objects that make up the robot
   */
  public PhysicalModel(List<ModelObject> modelObjects) {
    m_idMap = new HashMap<String, ModelBody>();
    m_allObjects = new ArrayList<ModelObject>();
    m_bodies = new ArrayList<ModelBody>();
    m_joints = new ArrayList<ModelJoint>();
    m_topObjects = new ArrayList<ModelObject>();
    
    m_listeners = Collections.synchronizedList(new ArrayList<ModelListener>());
    
    processModelObjects(modelObjects, true);
  }
  

  /**
   * @param model the model to set the libraries in
   * @param libraries the libraries to use to set the component libraries
   */
  public static void setLibraries(PhysicalModel model, ComponentLibraries libraries) {
    for (ModelBody body : model.getModelBodies()) {
      String libraryName = body.getLibraryName();
      if (!libraryName.isEmpty()) {
        ComponentLibrary library = libraries.getLibrary(libraryName);
        if (library == null) {
          System.err.println("Could not find library "+libraryName+" for body "+body.getId());
        }
        else
          body.setComponentLibrary(library);
      }
    }
  }

  /**
   * Add {@link ModelObject} to the {@link PhysicalModel}. The object added
   * is a top level objects. Children of the specified object will also be added.
   * @param modelObject the top level object to add
   */
  public void addTopLevelObject(ModelObject modelObject) {
    ArrayList<ModelObject> objs = new ArrayList<ModelObject>();
    objs.add(modelObject);
    processModelObjects(objs, true);
  }
  
  /**
   * Add {@link ModelObject}s to the {@link PhysicalModel}. The objects added
   * are NOT top level objects. Children of the specified objects will also be added.
   * This can be used when adding more children to existing {@link ModelObject}s in 
   * the {@link PhysicalModel}.
   * @param modelObject the component level objects to add
   */
  public void addComponentLevelObjects(List<ModelObject> modelObjects) {
    processModelObjects(modelObjects, false);
  }
  
  /**
   * Remove the specified object from the {@link PhysicalModel}
   * @param modelObject the model object to remove
   */
  public void removeObject(ModelObject modelObject) {
    if (modelObject instanceof ModelTransform) {
      ModelTransform modelTransform = (ModelTransform) modelObject;
      for (ModelObject target : modelTransform.getTransformTargets()) {
        removeObject(target);
      }
    }
    if (modelObject instanceof ModelJoint) {
      m_joints.remove((ModelJoint)modelObject);
    }
    if (modelObject instanceof ModelBody) {
      ModelBody modelBody = (ModelBody)modelObject;
      ModelObject parentObject = modelBody.getParentObject();
      if (parentObject instanceof ModelTransform)
    	  ((ModelTransform)parentObject).removeTransformTarget(modelBody);
      m_bodies.remove(modelBody);
      m_idMap.remove(modelBody.getId());
    }
    m_allObjects.remove(modelObject);
    m_topObjects.remove(modelObject);
  }
  
  /**
   * @return a copy of the physical model including copies of all of its components
   */
  public PhysicalModel copy() {
    // create a mapping of old to new objects
    Map<ModelObject,ModelObject> oldNewMap = new HashMap<ModelObject, ModelObject>();
    for (ModelObject obj : m_allObjects)
      oldNewMap.put(obj, obj.copy());
    
    // create a list of the new top objects, adding children recursively
    List<ModelObject> newTops = new ArrayList<ModelObject>();
    for (ModelObject topObj : m_topObjects)
      newTops.add(copyObject(topObj, oldNewMap));
    
    // create the new model
    return new PhysicalModel(newTops);
  }
  
  private ModelObject copyObject(ModelObject oldObj,
      Map<ModelObject, ModelObject> oldNewMap) {
    // recursively find the appropriate object and children in the map
    if (oldObj instanceof ModelTransform) {
      // a transform is the only tricky case -- we need to recurse to the children
      List<ModelObject> oldChildren = ((ModelTransform)oldObj).getTransformTargets();
      List<ModelObject> newChildren = new ArrayList<ModelObject>();
      for (ModelObject oldChild: oldChildren)
        newChildren.add(copyObject(oldChild, oldNewMap));
      ModelTransform newTransform = (ModelTransform) oldNewMap.get(oldObj);
      newTransform.addTransformTargets(newChildren);
      return newTransform;
    }
    else
      // just return the mapped object
      return oldNewMap.get(oldObj);
  }

  /**
   * Process a list of model objects updating the {@link PhysicalModel} data structures appropriately.
   * @param modelObjects the model objects to process
   * @param topLevel iff true add the objects in modelObjects to the list of top level objects
   * @return a list of the objects that were added to the model
   */
  private List<ModelObject> processModelObjects(List<ModelObject> modelObjects, boolean topLevel) {
    List<ModelObject> addedObjects = new ArrayList<ModelObject>();
    
    for (ModelObject modelObject : modelObjects) {
      m_allObjects.add(modelObject);
      if (topLevel)
        m_topObjects.add(modelObject);
      
      if (modelObject instanceof ModelBody) {
        ModelBody modelBody = (ModelBody) modelObject;
        m_bodies.add(modelBody);
        if (!m_idMap.containsKey(modelBody.getId()))
          m_idMap.put(modelBody.getId(), modelBody);
        else
          throw new RuntimeException("Duplicate ID not allowed!");
      }
      else if (modelObject instanceof ModelJoint) {
        m_joints.add((ModelJoint) modelObject);
      }
      else if (modelObject instanceof ModelTransform) {
        ModelTransform modelTransform = (ModelTransform) modelObject;
        addedObjects.addAll(processModelObjects(modelTransform.getTransformTargets(), false));
      }
      else {
        throw new RuntimeException("Unknown ModelObject subtype!");
      }
      addedObjects.add(modelObject);
    }
    return addedObjects;
  }
  
  /**
   * @return an attribute value view of the {@link PhysicalModel}
   */
  public AttributeValueView getAttributeValueView() {
    AttributeValueView attributeValueView = new AttributeValueView(this);
    for (ModelBody modelBody : m_bodies) {
      for (RobotAttributeValues rav : modelBody.getAttributeValues())
        attributeValueView.put(new RobotAttribute(rav.getAttribute()), rav.getMeasurement());
    }
    return attributeValueView;
  }
  
  public Map<String, Measurement> getAttributeMap() {
    HashMap<String, Measurement> attributes = new HashMap<String, Measurement>();
    for (ModelBody modelBody : m_bodies) {
      for (RobotAttributeValues rav : modelBody.getAttributeValues())
        attributes.put(rav.getAttribute(), rav.getMeasurement());
    }
    return attributes;
  }
  
  /**
   * Updates the {@link PhysicalModel} -- should only be called by a 
   * controller
   * @param rav the attribute / value to update
   * @return true iff found attribute to update
   */
  public boolean updateRobotAttributeFromDeltas(RobotAttributeValues rav) {
    String att = rav.getAttribute();
    String id = RobotAttributeValues.getPrefix(att);
    String postfix = RobotAttributeValues.getPostfix(att);
    
    ModelBody modelBody = m_idMap.get(id);
    if (modelBody != null) {
      if (!modelBody.isComponentFixed()) {
        // if the component for the body is fixed, do not update
        // TODO this is sort of hacky. We need a better way to do all of this
        List<ModelJoint> joints = getJoints(modelBody);
        modelBody.updateRobotAttributeFromDeltas((new RobotAttributeValues(postfix, rav.getMeasurement())), joints);
      }
      else
        // we found what we wanted to update, but we didn't change the model
        return true;
    }
    else
      //System.err.println("Could not find body "+id);
      return false;
    
    notifyModelChanged();
    return true;
  }
  
  /**
   * Get a list of joints for the specified body (body id used to find the appropriate joints)
   * @param modelBody the body to find the joints for
   * @return a list of joints attached to the specified body (based on id)
   */
  public List<ModelJoint> getJoints(ModelBody modelBody) {
    List<ModelJoint> joints = new ArrayList<ModelJoint>();
    String id = modelBody.getId();
    
    for (ModelJoint joint : m_joints) {
      if (joint.getBodyId1().equals(id) || joint.getBodyId2().equals(id))
        joints.add(joint);
    }
    
    return joints;
  }
  
  public List<ModelJoint> getModelJoints() {
	  return m_joints;
  }

  /**
   * Add a listener to be notified when the model has changed
   * @param listener the listener to add
   */
  public void addModelListener(ModelListener listener) {
    m_listeners.add(listener);
  }
  
  /**
   * Remove a listener to be notified when the model has changed
   * @param listener the listener to remove
   */
  public void removeModelListener(ModelListener listener) {
    m_listeners.remove(listener);
  }
  
  protected void notifyModelChanged() {
    List<ModelListener> copy = new ArrayList<ModelListener>(m_listeners);
    for (ModelListener listener : copy) {
      listener.modelChanged(new EventObject(this));
    }
  }
  
  /**
   * Write the robot model to the specified file as XML
   * @param file the file to write the XML to
   * @throws Exception if an error writing to the file occurs
   */
  public void writeToXMLFile(File file) throws Exception {
    ModelUtils.writeToXMLFile(file, m_topObjects);
  }

  /**
   * Create a {@link PhysicalModel} from the specified XML file
   * @param xmlFile the XML containing the robot
   * @param libraries the {@link ComponentLibraries} to link created objects to
   * @return a {@link PhysicalModel} created from the specified file
   * @throws InvalidModelException if the robot specified in the file is invalid
   */
  public static PhysicalModel loadModel(File xmlFile, ComponentLibraries libraries) throws InvalidModelException {
    DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
    
    try {
      // Using factory get an instance of document builder
      DocumentBuilder db = dbf.newDocumentBuilder();
      
      // Parse using builder to get DOM representation of the XML file
      Document dom = db.parse(xmlFile);
      
      Element rootElement = dom.getDocumentElement();
      if (!rootElement.getNodeName().equals("root")) {
        throw new InvalidModelException("Expected root node called \"root\", instead found: "+rootElement.getNodeName());
      }
      
      
      List<ModelObject> children = createChildrenModelObjects(rootElement);
      
      PhysicalModel physicalModel = new PhysicalModel(children);
      
      // set the libraries!
      PhysicalModel.setLibraries(physicalModel, libraries);
      
      return physicalModel;
    } catch(Exception e) {
      e.printStackTrace();
      throw new InvalidModelException(e.getMessage());
    }
  }

  private static List<ModelObject> createChildrenModelObjects(Element parentElement)
      throws InvalidModelException {
    List<ModelObject> m_children = new ArrayList<ModelObject>();
    
    // now parse the elements:
    NodeList children = parentElement.getChildNodes();
    for (int i=0; i<children.getLength(); i++) {
      Node node = children.item(i);
      if (node.getNodeType() == Node.ELEMENT_NODE) {
        Element element = (Element) node;
        String nodeName = element.getNodeName();
        if (nodeName.equals("body")) {
          ModelBody modelBody = ModelBody.createFromXML(element);
          m_children.add(modelBody);
        }
        else if (nodeName.equals("joint")) {
          ModelJoint modelJoint = ModelJoint.createFromXML(element);
          m_children.add(modelJoint);
        }
        else if (nodeName.equals("transform")) {
          ModelTransform modelTransform = ModelTransform.createFromXML(element);
          // now we need all the children that this affects
          List<ModelObject> transformChildren = createChildrenModelObjects(element);
          modelTransform.addTransformTargets(transformChildren);
          m_children.add(modelTransform);
        }
        else {
          throw new InvalidModelException("Unknown node name: "+nodeName);
        }
      }
    }
    
    // FIXME do I sort this list here?
    return m_children;
  }
  
  public Map<String, ModelBody> getIDMap() {
	  return m_idMap;
  }

  /**
   * @return all of the {@link ModelObject}s. These objects must NOT be modified.
   */
  public List<ModelObject> getModelObjects() {
    return m_allObjects;
  }

  public List<ModelBody> getModelBodies() {
    return m_bodies;
  }

  /**
   * @return a list of the top level {@link ModelObject}s. These objects must NOT be modified.
   */
  public List<ModelObject> getTopModelObjects() {
    return m_topObjects;
  }

}
