/* Parser for functional blueprints
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */

package com.bbn.morph.model.fb;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.bbn.morph.executer.Feature;
import com.bbn.morph.executer.Pair;
import com.bbn.morph.executer.StressInputType;
import com.bbn.morph.view.FB;
import com.bbn.morph.view.RobotAttribute;
import com.bbn.morph.view.StepDecrementUpdate;
import com.bbn.morph.view.StepIncrementUpdate;
import com.bbn.morph.view.StressFunction;
import com.bbn.morph.view.SymmetricFivePieceLinearStressFunction;
import com.bbn.morph.view.ThreePieceLinearStressFunction;
import com.bbn.morph.view.UpdateFunction;

/**
 * A class to parse an XML representation of functional blueprints into Java 
 * objects.
 * @author aadler
 *
 */
public class FBParser {

  /**
   * Parse the input stream (in XML format) into a list of {@link FB}s.
   * Functional blueprints are parsed iff there is no active attribute or it is set to true.
   * @param stream the stream of XML containing the functional blueprints
   * @return a list of functional blueprints
   * @throws InvalidFunctionalBlueprintException if the stream contains badly constructed functional blueprints or XML
   */
  public static List<FB> parseFB(InputStream stream) throws InvalidFunctionalBlueprintException {
    DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
    
    try {
      // Using factory get an instance of document builder
      DocumentBuilder db = dbf.newDocumentBuilder();
      
      // Parse using builder to get DOM representation of the XML stream
      Document dom = db.parse(stream);
      
      Element rootElement = dom.getDocumentElement();
      if (!rootElement.getNodeName().equals("functionalBlueprints")) {
        throw new InvalidFunctionalBlueprintException("Expected root node called \"blueprints\", instead found: "+rootElement.getNodeName());
      }
      
      
      List<FB> blueprints = new ArrayList<FB>();
      NodeList childNodes = rootElement.getChildNodes();
      for (int i=0; i<childNodes.getLength(); i++) {
        Node node = childNodes.item(i);
        if (node.getNodeType() == Node.ELEMENT_NODE) {
          Element element = (Element) node;
          FB fb = createBlueprint(element);
          if (fb != null)
            blueprints.add(fb);
        }
      }
      
      return blueprints;
    } catch(Exception e) {
      throw new InvalidFunctionalBlueprintException(e.getMessage());
    }
  }

  private static FB createBlueprint(Element element) throws InvalidFunctionalBlueprintException {
    String nodeName = element.getNodeName();
    if (nodeName != "functionalBlueprint")
      throw new InvalidFunctionalBlueprintException("Expected element called \"functionalBlueprint\", instead found: "+nodeName);
    // active if there is no active attribute or the attribute is true
    boolean isActive = (!element.hasAttribute("active") || 
        Boolean.parseBoolean(element.getAttribute("active")));
    if (isActive) {
      String name = element.getAttribute("name");
      if (name.isEmpty())
        throw new InvalidFunctionalBlueprintException("functionBlueprint node must have a non-empty \"name\" attribute.");
      
      NodeList stressNode = element.getElementsByTagName("stress");
      if (stressNode.getLength() != 1)
        throw new InvalidFunctionalBlueprintException("Expected exactly one \"stress\" node per functional blueprint.");
        
      NodeList updateNode = element.getElementsByTagName("updates");
      if (updateNode.getLength() != 1)
        throw new InvalidFunctionalBlueprintException("Expected exactly one \"updates\" node per functional blueprint.");
      
      StressFunction stressFunction = createStressFunction(stressNode.item(0));
      Map<RobotAttribute, UpdateFunction> updates = createUpdates(updateNode.item(0));
      
      return new FB(new Feature(name), updates, stressFunction);
    }
    else
      // if the functional blueprint is not active then we should just 
      // stop and return null
      return null;
  }

  private static Map<RobotAttribute, UpdateFunction> createUpdates(Node node) throws InvalidFunctionalBlueprintException {
    if (node.getNodeType() != Node.ELEMENT_NODE) 
      throw new InvalidFunctionalBlueprintException("\"updates\" node not Element as expected");

    Element updatesElement = (Element) node;
    NodeList updatesNodes = updatesElement.getElementsByTagName("update");
    
    Map<RobotAttribute, UpdateFunction> updateFunctions = new HashMap<RobotAttribute, UpdateFunction>();

    for (int i=0; i<updatesNodes.getLength(); i++) {
      Node updateNode = updatesNodes.item(i);
      if (updateNode.getNodeType() != Node.ELEMENT_NODE) 
        throw new InvalidFunctionalBlueprintException("\"update\" node not Element as expected");
     Element updateElement = (Element)updateNode;
     String attStr = updateElement.getAttribute("attribute");
     if (attStr.isEmpty())
       throw new InvalidFunctionalBlueprintException("update node must have a non-empty \"attribute\" attribute.");
     
     RobotAttribute attribute = new RobotAttribute(attStr);
     
     NodeList functionNodeList = updateElement.getElementsByTagName("function");
     if (functionNodeList.getLength() != 1)
       throw new InvalidFunctionalBlueprintException("Expected exactly one \"function\" node per udpate element.");
     Node functionNode = functionNodeList.item(0);
     UpdateFunction fn = createFunction(functionNode);

     updateFunctions.put(attribute, fn);
    }

    return updateFunctions;
  }

  private static UpdateFunction createFunction(Node functionNode) throws InvalidFunctionalBlueprintException {
    if (functionNode.getNodeType() != Node.ELEMENT_NODE) 
      throw new InvalidFunctionalBlueprintException("\"function\" node not Element as expected");

    Element functionElement = (Element) functionNode;
    String functionType = functionElement.getAttribute("type");
    
    NodeList valueNodes = functionElement.getElementsByTagName("value");
    if (valueNodes.getLength() != 1)
      throw new InvalidFunctionalBlueprintException("Expected exactly one \"value\" node per function element.");
    Node valueNode = valueNodes.item(0);
    if (valueNode.getNodeType() != Node.ELEMENT_NODE)
      throw new InvalidFunctionalBlueprintException("\"value\" node not Element as expected");
    Element valueElement = (Element)valueNode;
    String valueStr = valueElement.getTextContent();
    if (valueStr.isEmpty())
      throw new InvalidFunctionalBlueprintException("\"value\" node must contain a value.");
    double value = Double.parseDouble(valueStr);
    
    if (functionType.equals("stepDecrement")) {
      return new StepDecrementUpdate(value);
    }
    else if (functionType.equals("stepIncrement")) {
      return new StepIncrementUpdate(value);
    }
    else 
      throw new InvalidFunctionalBlueprintException("Unrecognized function type: "+functionType);
  }

  private static StressFunction createStressFunction(Node node) throws InvalidFunctionalBlueprintException {
    if (node.getNodeType() != Node.ELEMENT_NODE) 
      throw new InvalidFunctionalBlueprintException("\"stress\" node not Element as expected");

    Element element = (Element) node;
    String stressType = element.getAttribute("type");
    StressInputType stressInputType = StressInputType.fromString(stressType);
    
    NodeList functionNodes = element.getElementsByTagName("function");
    if (functionNodes.getLength() != 1)
      throw new InvalidFunctionalBlueprintException("Expected exactly one \"function\" node per stress element.");
    Node functionNode = functionNodes.item(0);
    if (functionNode.getNodeType() != Node.ELEMENT_NODE)
      throw new InvalidFunctionalBlueprintException("\"function\" node not Element as expected");
    Element functionElement = (Element) functionNode;
    String functionType = functionElement.getAttribute("type");

    if (functionType.equals("linearThreePiece")) {
      // first node:
      NodeList firstNodes = element.getElementsByTagName("first");
      if (firstNodes.getLength() != 1)
        throw new InvalidFunctionalBlueprintException("Expected exactly one \"first\" node per function element.");
      Node firstNode = firstNodes.item(0);
      if (firstNode.getNodeType() != Node.ELEMENT_NODE)
        throw new InvalidFunctionalBlueprintException("\"first\" node not Element as expected");
      Element firstElement = (Element)firstNode;
      if (!firstElement.hasAttribute("value"))
        throw new InvalidFunctionalBlueprintException("\"first\" node missing attribute \"value.\"");
      double firstValue = Double.parseDouble(firstElement.getAttribute("value"));
      if (!firstElement.hasAttribute("stress"))
        throw new InvalidFunctionalBlueprintException("\"first\" node missing attribute \"stress.\"");
      double firstStress = Double.parseDouble(firstElement.getAttribute("stress"));

      // second node:
      NodeList secondNodes = element.getElementsByTagName("second");
      if (secondNodes.getLength() != 1)
        throw new InvalidFunctionalBlueprintException("Expected exactly one \"second\" node per function element.");
      Node secondNode = secondNodes.item(0);
      if (secondNode.getNodeType() != Node.ELEMENT_NODE)
        throw new InvalidFunctionalBlueprintException("\"second\" node not Element as expected");
      Element secondElement = (Element)secondNode;
      if (!secondElement.hasAttribute("value"))
        throw new InvalidFunctionalBlueprintException("\"second\" node missing attribute \"value.\"");
      double secondValue = Double.parseDouble(secondElement.getAttribute("value"));
      if (!secondElement.hasAttribute("stress"))
        throw new InvalidFunctionalBlueprintException("\"second\" node missing attribute \"stress.\"");
      double secondStress = Double.parseDouble(secondElement.getAttribute("stress"));

      return new ThreePieceLinearStressFunction(stressInputType, 
          new Pair(firstValue, firstStress), new Pair(secondValue, secondStress));
    }

    else if (functionType.equals("linearSymmetricFivePiece")) {
      // mid node:
      NodeList midNodes = element.getElementsByTagName("mid");
      if (midNodes.getLength() != 1)
        throw new InvalidFunctionalBlueprintException("Expected exactly one \"mid\" node per function element.");
      Node midNode = midNodes.item(0);
      if (midNode.getNodeType() != Node.ELEMENT_NODE)
        throw new InvalidFunctionalBlueprintException("\"mid\" node not Element as expected");
      Element midElement = (Element)midNode;
      if (!midElement.hasAttribute("value"))
        throw new InvalidFunctionalBlueprintException("\"mid\" node missing attribute \"value.\"");
      double midValue = Double.parseDouble(midElement.getAttribute("value"));

      // innerRadius node:
      NodeList innerNodes = element.getElementsByTagName("innerRadius");
      if (innerNodes.getLength() != 1)
        throw new InvalidFunctionalBlueprintException("Expected exactly one \"innerRadius\" node per function element.");
      Node innerNode = innerNodes.item(0);
      if (innerNode.getNodeType() != Node.ELEMENT_NODE)
        throw new InvalidFunctionalBlueprintException("\"innerRadius\" node not Element as expected");
      Element innerElement = (Element)innerNode;
      if (!innerElement.hasAttribute("value"))
        throw new InvalidFunctionalBlueprintException("\"innerRadius\" node missing attribute \"value.\"");
      double innerRadiusValue = Double.parseDouble(innerElement.getAttribute("value"));

      // outerRadius node:
      NodeList outerNodes = element.getElementsByTagName("outerRadius");
      if (outerNodes.getLength() != 1)
        throw new InvalidFunctionalBlueprintException("Expected exactly one \"outerRadius\" node per function element.");
      Node outerNode = outerNodes.item(0);
      if (outerNode.getNodeType() != Node.ELEMENT_NODE)
        throw new InvalidFunctionalBlueprintException("\"outerRadius\" node not Element as expected");
      Element outerElement = (Element)outerNode;
      if (!outerElement.hasAttribute("value"))
        throw new InvalidFunctionalBlueprintException("\"outerRadius\" node missing attribute \"value.\"");
      double outerRadiusValue = Double.parseDouble(outerElement.getAttribute("value"));

      return new SymmetricFivePieceLinearStressFunction(stressInputType, 
          midValue, innerRadiusValue, outerRadiusValue);
    }

    else 
      throw new InvalidFunctionalBlueprintException("Unrecognized function type: "+functionType);
  }
  
}
