/* Test for functional blueprint parsing
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */

package com.bbn.morph.model.fb.test;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.List;
import java.util.Map;

import junit.framework.Assert;

import org.junit.Test;

import com.bbn.morph.executer.Pair;
import com.bbn.morph.executer.StressInputType;
import com.bbn.morph.model.fb.FBParser;
import com.bbn.morph.model.fb.InvalidFunctionalBlueprintException;
import com.bbn.morph.view.FB;
import com.bbn.morph.view.RobotAttribute;
import com.bbn.morph.view.StepDecrementUpdate;
import com.bbn.morph.view.StepIncrementUpdate;
import com.bbn.morph.view.StressFunction;
import com.bbn.morph.view.SymmetricFivePieceLinearStressFunction;
import com.bbn.morph.view.ThreePieceLinearStressFunction;
import com.bbn.morph.view.UpdateFunction;

public class FBParserTest {

  /**
   * Test that inactive blueprints are not processed.
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test
  public void inactiveTest() throws IOException, InvalidFunctionalBlueprintException {
    // test with active set to false
    String s = "<functionalBlueprints><functionalBlueprint name=\"Climb FB\" active=\"false\"/></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    List<FB> fbs = FBParser.parseFB(stream);
    Assert.assertTrue(fbs.isEmpty());
  }
  
  /**
   * Test for root node name
   * @throws InvalidFunctionalBlueprintException 
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void rootNodeTest() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprint/>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  
  /**
   * Test for functionBlueprint node name
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void blueprintNodeTest() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><blueprint/></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  
  /**
   * Test for functionBlueprint name attribute
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void blueprintNodeNameTest() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint/></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  
  /**
   * Test for functionBlueprint name attribute
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void blueprintNodeNameTest2() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint active=\"true\"/></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  
  /**
   * Test for functionBlueprint stress element test
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressNodeTest() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\"/></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint updates element test
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void updatesNodeTest() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\"><stress/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  
  /**
   * Test for functionBlueprint stress element test 1
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest1() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\"><stress/><updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  
  /**
   * Test for functionBlueprint stress element test 2
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest2() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"test\"/>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint stress element test 3
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest3() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"/>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint stress element test 4
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest4() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function/></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  
  /**
   * Test for functionBlueprint stress element test 5
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest5() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"test\"/></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  
  /**
   * Test for functionBlueprint stress element test 6
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest6() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\"/></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  
  /**
   * Test for functionBlueprint stress element test 7
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest7() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\"><first/>";
    s+="</function></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint stress element test 8
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest8() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\"><first value=\"5\"/>";
    s+="</function></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint stress element test 9
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest9() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\"><first value=\"5\" stress=\"5\"/>";
    s+="</function></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  
  /**
   * Test for functionBlueprint stress element test 10
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest10() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\"><first value=\"5\" stress=\"5\"/>";
    s+="<second/>";
    s+="</function></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint stress element test 11
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest11() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\"><first value=\"5\" stress=\"5\"/>";
    s+="<second value=\"10\"/>";
    s+="</function></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint stress element test 12 -- this one should parse
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test
  public void stressTest12() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\"><first value=\"5\" stress=\"5\"/>";
    s+="<second value=\"10\" stress=\"10\"/>";
    s+="</function></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  

  /**
   * Test for functionBlueprint stress element test 13
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest13() throws IOException, InvalidFunctionalBlueprintException {

    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearSymmetricFivePiece\">";
    s+="</function></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint stress element test 14
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest14() throws IOException, InvalidFunctionalBlueprintException {

    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearSymmetricFivePiece\">";
    s+="<mid/>";
    s+="</function></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint stress element test 15
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest15() throws IOException, InvalidFunctionalBlueprintException {

    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearSymmetricFivePiece\">";
    s+="<mid value=\"5\" />";
    s+="</function></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint stress element test 16
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest16() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearSymmetricFivePiece\">";
    s+="<mid value=\"5\" />";
    s+="<innerRadius/>";
    s+="</function></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint stress element test 17
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest17() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearSymmetricFivePiece\">";
    s+="<mid value=\"5\" />";
    s+="<innerRadius value=\"10\"/>";
    s+="</function></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint stress element test 18
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void stressTest18() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearSymmetricFivePiece\">";
    s+="<mid value=\"5\" />";
    s+="<innerRadius value=\"10\"/>";
    s+="<outerRadius/>";
    s+="</function></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint stress element test 19 -- should parse
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test
  public void stressTest19() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearSymmetricFivePiece\">";
    s+="<mid value=\"5\" />";
    s+="<innerRadius value=\"10\"/>";
    s+="<outerRadius value=\"3.6\"/>";
    s+="</function></stress>";
    s+="<updates/></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  
  /**
   * Test for functionBlueprint updates element test 1
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void updateTest1() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\">";
    s+="<first value=\"5\" stress=\"5\"/>";
    s+="<second value=\"10\" stress=\"10\"/>";
    s+="</function></stress>";
    s+="<updates>";
    s+="<update/>";
    s+="</updates></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint updates element test 2
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void updateTest2() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\">";
    s+="<first value=\"5\" stress=\"5\"/>";
    s+="<second value=\"10\" stress=\"10\"/>";
    s+="</function></stress>";
    s+="<updates>";
    s+="<update attribute=\"test\"/>";
    s+="</updates></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint updates element test 3
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void updateTest3() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\">";
    s+="<first value=\"5\" stress=\"5\"/>";
    s+="<second value=\"10\" stress=\"10\"/>";
    s+="</function></stress>";
    s+="<updates>";
    s+="<update attribute=\"test\">";
    s+="<function/>";
    s+="</update>";
    s+="</updates></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint updates element test 4
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void updateTest4() throws IOException, InvalidFunctionalBlueprintException {
    // test with active set to false
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\">";
    s+="<first value=\"5\" stress=\"5\"/>";
    s+="<second value=\"10\" stress=\"10\"/>";
    s+="</function></stress>";
    s+="<updates>";
    s+="<update attribute=\"test\">";
    s+="<function><value/></function>";
    s+="</update>";
    s+="</updates></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint updates element test 5
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void updateTest5() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\">";
    s+="<first value=\"5\" stress=\"5\"/>";
    s+="<second value=\"10\" stress=\"10\"/>";
    s+="</function></stress>";
    s+="<updates>";
    s+="<update attribute=\"test\">";
    s+="<function><value>5</value></function>";
    s+="</update>";
    s+="</updates></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  
  /**
   * Test for functionBlueprint updates element test 6
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test(expected=InvalidFunctionalBlueprintException.class)
  public void updateTest6() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\">";
    s+="<first value=\"5\" stress=\"5\"/>";
    s+="<second value=\"10\" stress=\"10\"/>";
    s+="</function></stress>";
    s+="<updates>";
    s+="<update attribute=\"test\">";
    s+="<function type=\"notvalid\"><value>5</value></function>";
    s+="</update>";
    s+="</updates></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  
  /**
   * Test for functionBlueprint updates element test 7 -- should work
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test
  public void updateTest7() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\">";
    s+="<first value=\"5\" stress=\"5\"/>";
    s+="<second value=\"10\" stress=\"10\"/>";
    s+="</function></stress>";
    s+="<updates>";
    s+="<update attribute=\"test\">";
    s+="<function type=\"stepDecrement\"><value>5</value></function>";
    s+="</update>";
    s+="</updates></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }

  /**
   * Test for functionBlueprint updates element test 8 -- should work
   * @throws IOException
   * @throws InvalidFunctionalBlueprintException
   */
  @Test
  public void updateTest8() throws IOException, InvalidFunctionalBlueprintException {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\">";
    s+="<first value=\"5\" stress=\"5\"/>";
    s+="<second value=\"10\" stress=\"10\"/>";
    s+="</function></stress>";
    s+="<updates>";
    s+="<update attribute=\"test\">";
    s+="<function type=\"stepIncrement\"><value>5</value></function>";
    s+="</update>";
    s+="</updates></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    FBParser.parseFB(stream);
  }
  
  ////// now to test some valid stuff:
  
  /**
   * Test for functionBlueprint -- full test
   * @throws Exception 
   */
  @Test
  public void fullTest() throws Exception {
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"time\"><function type=\"linearThreePiece\">";
    s+="<first value=\"3000\" stress=\"0\"/>";
    s+="<second value=\"5000\" stress=\"1000\"/>";
    s+="</function></stress>";
    s+="<updates>";
    s+="<update attribute=\"test1\">";
    s+="<function type=\"stepDecrement\"><value>2</value></function>";
    s+="</update>";
    s+="<update attribute=\"test2\">";
    s+="<function type=\"stepDecrement\"><value>2</value></function>";
    s+="</update>";
    s+="<update attribute=\"test3\">";
    s+="<function type=\"stepDecrement\"><value>1</value></function>";
    s+="</update>";
    s+="</updates></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    List<FB> fbs = FBParser.parseFB(stream);
    
    Assert.assertEquals(1, fbs.size());
    FB fb = fbs.get(0);
    Assert.assertEquals("test", fb.getFeature().getName());

    // look at the update functions
    Field field = fb.getClass().getDeclaredField("updateFunctions");
    field.setAccessible(true);
    @SuppressWarnings("unchecked")
    Map<RobotAttribute, UpdateFunction> map = (Map<RobotAttribute, UpdateFunction>) field.get(fb);
    Assert.assertEquals(3, map.size());
    
    UpdateFunction function = map.get(new RobotAttribute("test1"));
    Assert.assertTrue(function instanceof StepDecrementUpdate);
    Field functionField = function.getClass().getDeclaredField("step");
    functionField.setAccessible(true);
    Assert.assertEquals(2, functionField.getDouble(function), 0);
    
    function = map.get(new RobotAttribute("test2"));
    Assert.assertTrue(function instanceof StepDecrementUpdate);
    functionField = function.getClass().getDeclaredField("step");
    functionField.setAccessible(true);
    Assert.assertEquals(2, functionField.getDouble(function), 0);
    
    function = map.get(new RobotAttribute("test3"));
    Assert.assertTrue(function instanceof StepDecrementUpdate);
    functionField = function.getClass().getDeclaredField("step");
    functionField.setAccessible(true);
    Assert.assertEquals(1, functionField.getDouble(function), 0);
    
    // look at the stress function
    field = fb.getClass().getDeclaredField("stressF");
    field.setAccessible(true);
    StressFunction stressFunction = (StressFunction) field.get(fb);
    
    Assert.assertTrue(stressFunction instanceof ThreePieceLinearStressFunction);
    ThreePieceLinearStressFunction sf = (ThreePieceLinearStressFunction)stressFunction;
    
    field = sf.getClass().getDeclaredField("inputLabel");
    field.setAccessible(true);
    StressInputType sit = (StressInputType) field.get(sf);
    Assert.assertEquals("time", sit.toString().toLowerCase());
    
    field = sf.getClass().getDeclaredField("first");
    field.setAccessible(true);
    Pair p = (Pair) field.get(sf);
    Assert.assertEquals(3000, p.getX(), 0);
    Assert.assertEquals(0, p.getY(), 0);
    
    field = sf.getClass().getDeclaredField("second");
    field.setAccessible(true);
    p = (Pair) field.get(sf);
    Assert.assertEquals(5000, p.getX(), 0);
    Assert.assertEquals(1000, p.getY(), 0);
  }
  
  /**
   * Test for functionBlueprint -- full test
   * @throws Exception 
   */
  @Test
  public void fullTest2() throws Exception {
    double target = 3000;
    double r0 = 1000;
    double r1 = 2000;
    
    // test with active set to false
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"true\">";
    s+="<stress type=\"motor_mass\"><function type=\"linearSymmetricFivePiece\">";
    s+="<mid value=\""+target+"\"/>";
    s+="<innerRadius value=\""+r0+"\"/>";
    s+="<outerRadius value=\""+r1+"\"/>";
    s+="</function></stress>";
    s+="<updates>";
    s+="<update attribute=\"motor.mass\">";
    s+="<function type=\"stepIncrement\"><value>1.5</value></function>";
    s+="</update>";
    s+="</updates></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    List<FB> fbs = FBParser.parseFB(stream);
    
    Assert.assertEquals(1, fbs.size());
    FB fb = fbs.get(0);
    Assert.assertEquals("test", fb.getFeature().getName());

    // look at the update functions
    Field field = fb.getClass().getDeclaredField("updateFunctions");
    field.setAccessible(true);
    @SuppressWarnings("unchecked")
    Map<RobotAttribute, UpdateFunction> map = (Map<RobotAttribute, UpdateFunction>) field.get(fb);
    Assert.assertEquals(1, map.size());
    
    UpdateFunction function = map.get(new RobotAttribute("motor.mass"));
    Assert.assertTrue(function instanceof StepIncrementUpdate);
    Field functionField = function.getClass().getDeclaredField("step");
    functionField.setAccessible(true);
    Assert.assertEquals(1.5, functionField.getDouble(function), 0);
    
    // look at the stress function
    field = fb.getClass().getDeclaredField("stressF");
    field.setAccessible(true);
    StressFunction stressFunction = (StressFunction) field.get(fb);
    
    Assert.assertTrue(stressFunction instanceof SymmetricFivePieceLinearStressFunction);
    SymmetricFivePieceLinearStressFunction sf = (SymmetricFivePieceLinearStressFunction)stressFunction;
    
    field = sf.getClass().getDeclaredField("inputLabel");
    field.setAccessible(true);
    StressInputType sit = (StressInputType) field.get(sf);
    Assert.assertEquals("motor_mass", sit.toString().toLowerCase());
    
    field = sf.getClass().getDeclaredField("r0");
    field.setAccessible(true);
    Assert.assertEquals(r0, field.getDouble(sf), 0);

    field = sf.getClass().getDeclaredField("r1");
    field.setAccessible(true);
    Assert.assertEquals(r1, field.getDouble(sf), 0);
    
    // need to check left and right three piece functions:
    field = sf.getClass().getDeclaredField("left");
    field.setAccessible(true);
    StressFunction leftSF = (StressFunction) field.get(sf);
    Assert.assertTrue(leftSF instanceof ThreePieceLinearStressFunction);
    ThreePieceLinearStressFunction left = (ThreePieceLinearStressFunction)leftSF;
    field = left.getClass().getDeclaredField("inputLabel");
    field.setAccessible(true);
    sit = (StressInputType) field.get(left);
    Assert.assertEquals("motor_mass", sit.toString().toLowerCase());
    field = left.getClass().getDeclaredField("first");
    field.setAccessible(true);
    Pair p = (Pair) field.get(left);
    Assert.assertEquals(target-r1, p.getX(), 0);
    Assert.assertEquals(1, p.getY(), 0);
    field = left.getClass().getDeclaredField("second");
    field.setAccessible(true);
    p = (Pair) field.get(left);
    Assert.assertEquals(target-r0, p.getX(), 0);
    Assert.assertEquals(0, p.getY(), 0);
    
    field = sf.getClass().getDeclaredField("right");
    field.setAccessible(true);
    StressFunction rightSF = (StressFunction) field.get(sf);
    Assert.assertTrue(rightSF instanceof ThreePieceLinearStressFunction);
    ThreePieceLinearStressFunction right = (ThreePieceLinearStressFunction)rightSF;
    field = right.getClass().getDeclaredField("inputLabel");
    field.setAccessible(true);
    sit = (StressInputType) field.get(right);
    Assert.assertEquals("motor_mass", sit.toString().toLowerCase());
    field = right.getClass().getDeclaredField("first");
    field.setAccessible(true);
    p = (Pair) field.get(right);
    Assert.assertEquals(target+r0, p.getX(), 0);
    Assert.assertEquals(0, p.getY(), 0);
    field = right.getClass().getDeclaredField("second");
    field.setAccessible(true);
    p = (Pair) field.get(right);
    Assert.assertEquals(target+r1, p.getX(), 0);
    Assert.assertEquals(1, p.getY(), 0);
  }

  /**
   * Test for functionBlueprint -- full test, but no FBs
   * @throws Exception 
   */
  @Test
  public void fullTest3() throws Exception {
    // test with active set to false
    String s = "<functionalBlueprints><functionalBlueprint name=\"test\" active=\"false\">";
    s+="<stress type=\"motor.mass\"><function type=\"linearSymmetricFivePiece\">";
    s+="<mid value=\"3000\"/>";
    s+="<innerRadius value=\"1000\"/>";
    s+="<outerRadius value=\"2000\"/>";
    s+="</function></stress>";
    s+="<updates>";
    s+="<update attribute=\"motor.mass\">";
    s+="<function type=\"stepIncrement\"><value>1.5</value></function>";
    s+="</update>";
    s+="</updates></functionalBlueprint></functionalBlueprints>";
    ByteArrayInputStream stream = new ByteArrayInputStream(s.getBytes("UTF-8"));
    
    List<FB> fbs = FBParser.parseFB(stream);
    
    Assert.assertEquals(0, fbs.size());
  }
}
