/* Interface to running (and displaying) Proto
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.ui;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Scanner;

import com.bbn.morph.model.CompositeModel;
import com.bbn.morph.view.ProtoModelView;

/**
 * A class to spawn a simulation of the current robot and environment. 
 * For display purposes only -- doesn't return any results.
 * @author aadler
 *
 */
public class ProtoSimulator {

  public static final String defaultFileName = "resources/xmlModels/autoGenProtoUI.xml";
  private CompositeModel m_model;


  public ProtoSimulator(CompositeModel model) {
    m_model = model;
  }

  public void  attemptProtoCall() {
    ProtoModelView cmv = new ProtoModelView(m_model.getRobotModel(), 
        m_model.getEnvironmentModel());
    File file = new File(defaultFileName);
    try {
      cmv.writeToFile(file);
    }
    catch (Exception e) {
      System.err.println(e.toString());
      return;
    }
    
    boolean processEndDetected = false;

    Runtime runtime = Runtime.getRuntime();
    int bodies = cmv.getNumberOfBodies();

    String[] cmd = {"/usr/local/bin/proto", "-L", "odedynamics", "-body", 
        file.getAbsolutePath(), 
        "-n", Integer.toString(bodies), "-m", "-dim", "10", "10", "-3d", 
        "-floor", "-gravity", "-10","-l",
        "-s", "0.1", "(all (red (bump)) (madvreport (bump) 0))", "-L", "madv",
        "-stop-after", "50", "-T", "--end-x", Double.toString(cmv.getCenterOfStep().getStandardValue())};
    //"--DLL", "libmadvplugin.la", 


    try {
      Process exec = runtime.exec(cmd);
      int pingTime = 500;

      // monitor the streams from the process
      StreamMonitorThread inputStreamThread = new StreamMonitorThread(exec.getInputStream(), Thread.currentThread());
      inputStreamThread.start();
      ErrorStreamMonitorThread errorStreamThread = new ErrorStreamMonitorThread(exec.getErrorStream(), System.err, Thread.currentThread());
      errorStreamThread.start();

      while (!processEndDetected) {
        boolean stillRunning = isProcessRunning(exec);
        if (!stillRunning) {
          processEndDetected = true;
          // we monitor the main thread and it has exited
          System.out.println("Process killed.\n");
          inputStreamThread.interrupt();
          errorStreamThread.interrupt();
          exec.destroy();
        }
        else {
          try {
            // the main thread is still running, so wait and then check again
            Thread.sleep(pingTime);
          }
          catch (InterruptedException ie) {
            // here we just want to go and immediately check for a crash
            if (errorStreamThread.isDone() || inputStreamThread.isDone() ){
              inputStreamThread.interrupt();
              errorStreamThread.interrupt();
              exec.destroy();
              processEndDetected=true;
            }

          }
        }
      }
    }
    catch (IOException e) {
      System.err.println(e.toString());
      return;
    }
  }

  /**
   * @param p the {@link Process} to check to see if it is still running
   * @return true iff the process is still running
   */
  private boolean isProcessRunning(Process p) {
    try {
      p.exitValue();
      return false;
    } catch (IllegalThreadStateException e) {
      return true;
    }
  }


  private class ErrorStreamMonitorThread extends Thread {
    private InputStream m_inputStream;
    private OutputStream m_outputStream;
    private Thread m_thread;
    private boolean m_done = false;

    public ErrorStreamMonitorThread(InputStream inputStream, OutputStream outputStream, Thread thread) {
      m_inputStream = inputStream;
      m_outputStream = outputStream;
      m_thread = thread;
    }

    @Override
    public void run() {
      // FIXME this is not efficient
      int b;
      try {
        while ((b = m_inputStream.read()) != -1) {
          m_outputStream.write(b);
        }
      }
      catch (IOException e) {
        System.err.println("Error stream monitor exception"+e.toString());
      }

      // if we run out of input, then we should flush and interrupt the parent 
      // thread
      m_done =true;
      m_thread.interrupt();
    }

    public boolean isDone(){
      return m_done;
    }
  }

  private class StreamMonitorThread extends Thread {
    private InputStream m_inputStream;
    private Thread m_thread;
    private boolean m_done = false;

    public StreamMonitorThread(InputStream inputStream, Thread thread) {
      m_inputStream = inputStream;
      m_thread = thread;
    }

    @Override
    public void run() {
      Scanner scanner = new Scanner(m_inputStream);
      // Look for the finished message:
      scanner.findWithinHorizon("Finished climbing at:", 0);

      // if we run out of input, then we should flush and interrupt the parent 
      // thread
      m_done = true;
      m_thread.interrupt();
    }
    
    public boolean isDone(){
      return m_done;
    }
  }

}
