/*
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.view;

import java.io.StringWriter;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.bbn.morph.model.Measurement;
import com.bbn.morph.model.ModelBody;
import com.bbn.morph.model.PhysicalModel;
import com.bbn.morph.model.Units;

public class ROSModelView {

  private PhysicalModel m_robot;
  private PhysicalModel m_environment;

  // constants that are canned at least for now:
  private Measurement m_wheelLength = new Measurement(Units.getMeters(), 0.0254);
  private Measurement m_treadFactor = new Measurement(Units.getNone(), 0.5);
  private Measurement m_wheelBuffer = new Measurement(Units.getMeters(), 0.0025);
  
  private Measurement m_wheelRadius = new Measurement(Units.getMeters(), 0.027940); // FIXME this used to be calculated from other params!
  private Measurement m_idleWheelOffset = new Measurement(Units.getMeters(), 0.019050);

  // FIXME static pid values?
  private static final double P = 1.0;
  private static final double I = 20.0;
  private static final double I_CLAMP = 100.0;
  
  
  /**
   * Construct a new {@link ROSModelView}
   * @param robot the robot model
   * @param environment the environment model
   */
  public ROSModelView(PhysicalModel robot, PhysicalModel environment) {
    m_robot = robot;
    m_environment = environment;
  }
  
  public String[] getFiles() {
    int numWheels = calculateNumWheels();
    
    String[] files = new String[5];
    
    try {
      files[0] = getDimensionsFile();
      files[1] = getWheelFile(numWheels);
      files[2] = getWheelControllerFile(numWheels, P, I, I_CLAMP);
      files[3] = getControllersLaunchFiles(numWheels);
      files[4] = getStepModelFile();
      
      return files;
    }
    catch (Exception e) {
      e.printStackTrace();
    }
    
    return new String[0];
  }

  private Document createDocument() throws ParserConfigurationException {
    DocumentBuilderFactory documentFactory = DocumentBuilderFactory.newInstance();
    DocumentBuilder documentBuilder = documentFactory.newDocumentBuilder();
 
    //root elements
    Document doc = documentBuilder.newDocument();
    doc.setXmlVersion("1.0");
    
    return doc;
  }
  
  private String convertDocumentToString(Document doc) throws TransformerException {
    // set up a transformer:
    Transformer transformer = TransformerFactory.newInstance().newTransformer();
    transformer.setOutputProperty(OutputKeys.INDENT, "yes");
    
    transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
    
    // write out to a file
    DOMSource source = new DOMSource(doc);
    StringWriter sw = new StringWriter();
    StreamResult result = new StreamResult(sw);
    transformer.transform(source, result);
    return sw.toString();
  }
  
  /**
   * @return the number of wheels in the tread
   */
  private int calculateNumWheels() {
    Map<String, ModelBody> idMap = m_robot.getIDMap();
    ModelBody botBody = idMap.get("botbody");
    Measurement distance = botBody.getAttributeValue("lengthx").subtract(m_idleWheelOffset);
    double wheels = distance.divide(m_wheelRadius.mult(m_treadFactor)).getStandardValue();
    wheels = Math.floor(wheels)+1;
    return (int)wheels;
  }
  
  private String getControllersLaunchFiles(int numOfWheels) throws ParserConfigurationException, TransformerException {
    // first generate the string of controllers:
    StringBuilder controllers = new StringBuilder();
    
    // unique controller:
    controllers.append("lf_controller");
    
    for (int j=0; j<numOfWheels; j++) {
      controllers.append(" r"+j+"_controller");
      controllers.append(" l"+j+"_controller");
    }

    // now generate the file/string
    Document doc = createDocument();
    
    Element launch = doc.createElement("launch");
    launch.appendChild(doc.createComment("Start controllers"));
    
    Element node = doc.createElement("node");
    node.setAttribute("name", "default_controllers_spawner");
    node.setAttribute("pkg", "pr2_controller_manager");
    node.setAttribute("type", "spawner");
    node.setAttribute("output", "screen");
    node.setAttribute("args", controllers.toString());

    launch.appendChild(node);
    doc.appendChild(launch);

    return convertDocumentToString(doc);
  }
  
  private String getWheelControllerFile(int numOfWheels, double p, double i, double iClamp) {
    StringBuilder sb = new StringBuilder();
    
    sb.append("num_wheels : "+Integer.toString(numOfWheels)+"\n");
    sb.append("\n");
    
    for (int j=0; j<numOfWheels; j++) {
      sb.append("r"+j+"_controller:\n");
      sb.append("   type: robot_mechanism_controllers/JointVelocityController\n");
      sb.append("   joint: base_to_right_"+j+"_wheel\n");
      outputPID(sb, p, i, iClamp);
      sb.append("\n");

      sb.append("l"+j+"_controller:\n");
      sb.append("   type: robot_mechanism_controllers/JointVelocityController\n");
      sb.append("   joint: base_to_left_"+j+"_wheel\n");
      outputPID(sb, p, i, iClamp);
      sb.append("\n");
    }
    
    return sb.toString();
  }

  private void outputPID(StringBuilder sb, double p, double i, double iClamp) {
    sb.append("   pid:\n");
    sb.append("     p: " + Double.toString(p) + "\n");
    sb.append("     i: " + Double.toString(i) + "\n");
    sb.append("     i_clamp: " + Double.toString(iClamp) + "\n");
  }
  

  
  private String getDimensionsFile() throws ParserConfigurationException, TransformerException {
    Document doc = createDocument();
    
    Element robot = doc.createElement("robot");
    robot.setAttribute("xmlns:xi", "http://www.w3.org/2001/XInclude");
    robot.setAttribute("xmlns:gazebo", "http://playerstage.sourceforge.net/gazebo/xmlschema/#gz");
    robot.setAttribute("xmlns:model", "http://playerstage.sourceforge.net/gazebo/xmlschema/#model");
    robot.setAttribute("xmlns:sensor", "http://playerstage.sourceforge.net/gazebo/xmlschema/#sensor");
    robot.setAttribute("xmlns:body", "http://playerstage.sourceforge.net/gazebo/xmlschema/#body");
    robot.setAttribute("xmlns:geom", "http://playerstage.sourceforge.net/gazebo/xmlschema/#geom");
    robot.setAttribute("xmlns:joint", "http://playerstage.sourceforge.net/gazebo/xmlschema/#joint");
    robot.setAttribute("xmlns:interface", "http://playerstage.sourceforge.net/gazebo/xmlschema/#interface");
    robot.setAttribute("xmlns:rendering", "http://playerstage.sourceforge.net/gazebo/xmlschema/#rendering");
    robot.setAttribute("xmlns:renderable", "http://playerstage.sourceforge.net/gazebo/xmlschema/#renderable");
    robot.setAttribute("xmlns:controller", "http://playerstage.sourceforge.net/gazebo/xmlschema/#controller");
    robot.setAttribute("xmlns:physics", "http://playerstage.sourceforge.net/gazebo/xmlschema/#physics");
    robot.setAttribute("xmlns:xacro", "http://playerstage.sourceforge.net/gazebo/xmlschema/#interface");
    robot.setAttribute("name", "minidroid");

    Map<String, ModelBody> idMap = m_robot.getIDMap();
    ModelBody botBody = idMap.get("botbody");
    ModelBody leftFlipper = idMap.get("leftarm");
    ModelBody rightFlipper = idMap.get("rightarm");

    
    robot.appendChild(doc.createComment("body dimensions"));
    Element bodyWidth = doc.createElement("xacro:property");
    bodyWidth.setAttribute("name", "body_width");
    bodyWidth.setAttribute("value", Double.toString(botBody.getAttributeValue("lengthy").getStandardValue())); //"0.12090");
    robot.appendChild(bodyWidth);
    Element bodyLength = doc.createElement("xacro:property");
    bodyLength.setAttribute("name", "body_length");
    bodyLength.setAttribute("value", Double.toString(botBody.getAttributeValue("lengthx").getStandardValue())); //"0.17780");
    robot.appendChild(bodyLength);
    Element bodyHeight = doc.createElement("xacro:property");
    bodyHeight.setAttribute("name", "body_height");
    bodyHeight.setAttribute("value", Double.toString(botBody.getAttributeValue("lengthz").getStandardValue())); //"0.043180"); 
    robot.appendChild(bodyHeight);
    
    robot.appendChild(doc.createComment("wheel dimensions"));
    Element wheelRadius = doc.createElement("xacro:property");
    wheelRadius.setAttribute("name", "wheel_radius");
    wheelRadius.setAttribute("value", Double.toString(m_wheelRadius.getStandardValue()));
    robot.appendChild(wheelRadius);
    Element wheelLength = doc.createElement("xacro:property");
    wheelLength.setAttribute("name", "wheel_length");
    wheelLength.setAttribute("value", Double.toString(m_wheelLength.getStandardValue()));
    robot.appendChild(wheelLength);
    
    robot.appendChild(doc.createComment("flipper dimensions"));
    Element lFlipperWidth = doc.createElement("xacro:property");
    lFlipperWidth.setAttribute("name", "left_flipper_width");
    lFlipperWidth.setAttribute("value", Double.toString(leftFlipper.getAttributeValue("lengthy").getStandardValue())); //"0.0079375");
    robot.appendChild(lFlipperWidth);
    Element lFlipperLength = doc.createElement("xacro:property");
    lFlipperLength.setAttribute("name", "left_flipper_length");
    lFlipperLength.setAttribute("value", Double.toString(leftFlipper.getAttributeValue("lengthz").getStandardValue())); //"0.13970");
    robot.appendChild(lFlipperLength);
    Element lFlipperHeight = doc.createElement("xacro:property");
    lFlipperHeight.setAttribute("name", "left_flipper_height");
    lFlipperHeight.setAttribute("value", Double.toString(leftFlipper.getAttributeValue("lengthx").getStandardValue())); //"0.0254");
    robot.appendChild(lFlipperHeight);
    Element rFlipperWidth = doc.createElement("xacro:property");
    rFlipperWidth.setAttribute("name", "right_flipper_width");
    rFlipperWidth.setAttribute("value", Double.toString(rightFlipper.getAttributeValue("lengthy").getStandardValue())); //"0.0079375");
    robot.appendChild(rFlipperWidth);
    Element rFlipperLength = doc.createElement("xacro:property");
    rFlipperLength.setAttribute("name", "right_flipper_length");
    rFlipperLength.setAttribute("value", Double.toString(rightFlipper.getAttributeValue("lengthz").getStandardValue())); //"0.13970");
    robot.appendChild(rFlipperLength);
    Element rFlipperHeight = doc.createElement("xacro:property");
    rFlipperHeight.setAttribute("name", "right_flipper_height");
    rFlipperHeight.setAttribute("value", Double.toString(rightFlipper.getAttributeValue("lengthx").getStandardValue())); //"0.0254");
    robot.appendChild(rFlipperHeight);
    
    robot.appendChild(doc.createComment("wheel offsets"));
    Element wheelBuffer = doc.createElement("xacro:property");
    wheelBuffer.setAttribute("name", "wheel_buffer");
    wheelBuffer.setAttribute("value", Double.toString(m_wheelBuffer.getStandardValue()));
    robot.appendChild(wheelBuffer);
    Element idleWheelOffset = doc.createElement("xacro:property");
    idleWheelOffset.setAttribute("name", "idle_wheel_offset");
    idleWheelOffset.setAttribute("value", Double.toString(m_idleWheelOffset.getStandardValue()));
    robot.appendChild(idleWheelOffset);
 
    robot.appendChild(doc.createComment("axle locations"));
    Element idleAxle = doc.createElement("xacro:property");
    idleAxle.setAttribute("name", "idle_axle");
    idleAxle.setAttribute("value", "${(-body_length/2) + wheel_radius - idle_wheel_offset}"); // FIXME not sure it's necessary
    robot.appendChild(idleAxle);
    
    robot.appendChild(doc.createComment("tread spacing"));
    Element treadSpace = doc.createElement("xacro:property");
    treadSpace.setAttribute("name", "tread_factor");
    treadSpace.setAttribute("value", Double.toString(m_treadFactor.getStandardValue()));
    robot.appendChild(treadSpace);
 
    doc.appendChild(robot);
    
    return convertDocumentToString(doc);
  }
  
  
  private String getWheelFile(int numberOfWheels) throws ParserConfigurationException, TransformerException {
    Document doc = createDocument();
    
    Element robot = doc.createElement("robot");
    robot.setAttribute("xmlns:xi", "http://www.w3.org/2001/XInclude");
    robot.setAttribute("xmlns:gazebo", "http://playerstage.sourceforge.net/gazebo/xmlschema/#gz");
    robot.setAttribute("xmlns:model", "http://playerstage.sourceforge.net/gazebo/xmlschema/#model");
    robot.setAttribute("xmlns:sensor", "http://playerstage.sourceforge.net/gazebo/xmlschema/#sensor");
    robot.setAttribute("xmlns:body", "http://playerstage.sourceforge.net/gazebo/xmlschema/#body");
    robot.setAttribute("xmlns:geom", "http://playerstage.sourceforge.net/gazebo/xmlschema/#geom");
    robot.setAttribute("xmlns:joint", "http://playerstage.sourceforge.net/gazebo/xmlschema/#joint");
    robot.setAttribute("xmlns:interface", "http://playerstage.sourceforge.net/gazebo/xmlschema/#interface");
    robot.setAttribute("xmlns:rendering", "http://playerstage.sourceforge.net/gazebo/xmlschema/#rendering");
    robot.setAttribute("xmlns:renderable", "http://playerstage.sourceforge.net/gazebo/xmlschema/#renderable");
    robot.setAttribute("xmlns:controller", "http://playerstage.sourceforge.net/gazebo/xmlschema/#controller");
    robot.setAttribute("xmlns:physics", "http://playerstage.sourceforge.net/gazebo/xmlschema/#physics");
    robot.setAttribute("xmlns:xacro", "http://playerstage.sourceforge.net/gazebo/xmlschema/#interface");
    robot.setAttribute("name", "minidroid");

    for (int i=0; i < numberOfWheels; i++) {
      Element wheel = doc.createElement("xacro:wheel_axle");
      wheel.setAttribute("wheel_index", Integer.toString(i));
      robot.appendChild(wheel);
    }

    doc.appendChild(robot);
    
    return convertDocumentToString(doc);
  }
    
  
  private String getStepModelFile() throws ParserConfigurationException, TransformerException {
    Document doc = createDocument();
    
    Element model = doc.createElement("model:physical");
    model.setAttribute("xmlns:xi", "http://www.w3.org/2001/XInclude");
    model.setAttribute("xmlns:gazebo", "http://playerstage.sourceforge.net/gazebo/xmlschema/#gz");
    model.setAttribute("xmlns:model", "http://playerstage.sourceforge.net/gazebo/xmlschema/#model");
    model.setAttribute("xmlns:sensor", "http://playerstage.sourceforge.net/gazebo/xmlschema/#sensor");
    model.setAttribute("xmlns:window", "http://playerstage.sourceforge.net/gazebo/xmlschema/#window");
    model.setAttribute("xmlns:param", "http://playerstage.sourceforge.net/gazebo/xmlschema/#param");
    model.setAttribute("xmlns:body", "http://playerstage.sourceforge.net/gazebo/xmlschema/#body");
    model.setAttribute("xmlns:geo", "http://willowgarage.com/xmlschema/#geo");
    model.setAttribute("xmlns:geom", "http://playerstage.sourceforge.net/gazebo/xmlschema/#geom");
    model.setAttribute("xmlns:joint", "http://playerstage.sourceforge.net/gazebo/xmlschema/#joint");
    model.setAttribute("xmlns:interface", "http://playerstage.sourceforge.net/gazebo/xmlschema/#interface");
    model.setAttribute("xmlns:ui", "http://playerstage.sourceforge.net/gazebo/xmlschema/#ui");
    model.setAttribute("xmlns:rendering", "http://playerstage.sourceforge.net/gazebo/xmlschema/#rendering");
    model.setAttribute("xmlns:renderable", "http://playerstage.sourceforge.net/gazebo/xmlschema/#renderable");
    model.setAttribute("xmlns:controller", "http://playerstage.sourceforge.net/gazebo/xmlschema/#controller");
    model.setAttribute("xmlns:physics", "http://playerstage.sourceforge.net/gazebo/xmlschema/#physics");
    model.setAttribute("name", "stair");
    
    Map<String, ModelBody> idMap = m_environment.getIDMap();
    ModelBody step = idMap.get("step");
    double stepLength = step.getAttributeValue("lengthx").getStandardValue();
    double stepWidth = step.getAttributeValue("lengthy").getStandardValue();
    double stepHeight = step.getAttributeValue("lengthz").getStandardValue();

    
    // FIXME these values should come from the step position as specified in the object
    Element xyz = doc.createElement("xyz");
    xyz.setTextContent("0 "+Double.toString(-stepWidth/2)+" "+Double.toString(stepHeight/2));
    model.appendChild(xyz);
    
    Element rpy = doc.createElement("rpy");
    rpy.setTextContent("0.0 0.0 0.0");
    model.appendChild(rpy);
    
    Element staticE = doc.createElement("static");
    staticE.setTextContent("true");
    model.appendChild(staticE);
    
    Element box = doc.createElement("body:box");
    box.setAttribute("name", "stair");
    Element geom = doc.createElement("geom:box");
    geom.setAttribute("name", "wall_1_geom");
    Element mesh = doc.createElement("mesh");
    mesh.setTextContent("default");
    geom.appendChild(mesh);
    Element size = doc.createElement("size");
    size.setTextContent(Double.toString(stepLength)+" "+Double.toString(stepWidth)+" "+Double.toString(stepHeight));
    geom.appendChild(size);
    Element visual = doc.createElement("visual");
    Element size2 = doc.createElement("size");
    size2.setTextContent(Double.toString(stepLength)+" "+Double.toString(stepWidth)+" "+Double.toString(stepHeight));
    visual.appendChild(size2);
    Element material = doc.createElement("material");
    material.setTextContent("Gazebo/Rocky");
    visual.appendChild(material);
    Element mesh2 = doc.createElement("mesh");
    mesh2.setTextContent("unit_box");
    visual.appendChild(mesh2);
    geom.appendChild(visual);
    box.appendChild(geom);
    model.appendChild(box);
    
    doc.appendChild(model);
    
    return convertDocumentToString(doc);
  }
  
}
