/*
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.view;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import com.bbn.morph.model.Measurement;
import com.bbn.morph.model.ModelBody;
import com.bbn.morph.model.ModelCylinder;
import com.bbn.morph.model.ModelHingeJoint;
import com.bbn.morph.model.ModelObject;
import com.bbn.morph.model.ModelPosition;
import com.bbn.morph.model.ModelQuaternion;
import com.bbn.morph.model.ModelRotation;
import com.bbn.morph.model.ModelTransform;
import com.bbn.morph.model.PhysicalModel;
import com.bbn.morph.model.RobotAttributeValues;
import com.bbn.morph.model.Units;

/**
 * A view that combines the robot and environment into one view
 * @author aadler
 *
 */
public abstract class SimulatorModelView {

  protected PhysicalModel m_modifiedRobot;
  protected PhysicalModel m_environment;

  // wheel constants:
  protected Measurement m_wheelDistanceFromPassiveEdge = new Measurement(Units.getMeters(), 0.5);
  protected Measurement m_wheelDistanceFromDriveEdge = new Measurement(Units.getMeters(), 0.5);
  //private static double WHEEL_RADIUS = 0.75;
  protected Measurement m_wheelLength = new Measurement(Units.getMeters(), 0.5);
  protected Measurement m_wheelGap = new Measurement(Units.getMeters(), 0.05); // gap between wheel and body
  protected Measurement m_wheelOverlap = new Measurement(Units.getMeters(), 0.125); // amount of wheel hanging over the corner
  protected Measurement m_wheelMass = new Measurement(Units.getKilograms(), 1.0);
  
  protected Measurement m_flipperGap = new Measurement(Units.getMeters(), 0.1); // gap between wheel and flipper
  
  protected Measurement m_stepCenterDelta = new Measurement(Units.getMeters(), 1.0);

  /**
   * Construct a new {@link SimulatorModelView}
   * @param robot the robot model
   * @param environment the environment model
   */
  public SimulatorModelView(PhysicalModel robot, PhysicalModel environment) {
    setVariables();
    m_modifiedRobot = modifyRobot(robot.copy());
    m_environment = environment;
  }

  /**
   * Construct a new {@link SimulatorModelView} with a step of a certain height.
   * Useful for testing
   * @param robot the robot model
   * @param environment the environment model
   * @param stepHeight the height of the step to make
   */
  public SimulatorModelView(PhysicalModel robot, PhysicalModel environment, Measurement stepHeight) {
    setVariables();
    m_modifiedRobot = modifyRobot(robot.copy());
    m_environment = modifyEnvironment(environment.copy(), stepHeight);
  }
  
  /**
   * Set any variables that are necessary
   */
  protected void setVariables() {
    // nothing to do by default
  }


  public int getNumberOfBodies() {
    return m_modifiedRobot.getModelBodies().size() + m_environment.getModelBodies().size();
  }
  
  public Measurement getCenterOfStep() {
    return (m_environment.getAttributeMap().get("step.x")).add(m_stepCenterDelta);
  }
  
  // modify the robot by adding wheels and moving the flippers away from the body a bit
  protected PhysicalModel modifyRobot(PhysicalModel copy) {
    Map<String, ModelBody> idMap = copy.getIDMap();
    ModelBody botBody = idMap.get("botbody");
    Measurement botLength = botBody.getAttributeValue("lengthx");
    Measurement botWidth = botBody.getAttributeValue("lengthy");
    Measurement botHeight = botBody.getAttributeValue("lengthz");
    ModelBody motor = idMap.get("motor");
    copy.removeObject(motor);
    
    // calcualte the wheel radius:
    Measurement maxDistance = Measurement.max(m_wheelDistanceFromDriveEdge, m_wheelDistanceFromPassiveEdge);
    Measurement wheelRadius = (maxDistance.mult(maxDistance)).add((botHeight.divideMD(2.0)).mult(botHeight.divideMD(2.0))).sqrt().add(m_wheelOverlap);
    //takes the sqrt of (wheel distance from edge^2 plus (botheight/2)^2) plus Wheel Overlap
    
    // body and existing axis should be at this height
    Measurement botZPosition = Measurement.max(botHeight.divideMD(2.0), wheelRadius);
    Measurement zDelta = botZPosition.subtract(botBody.getAttributeValue("z"));
    botBody.updateRobotAttributeFromDeltas(new RobotAttributeValues("z", zDelta), copy.getJoints(botBody));
    
    
    
    // we should shift all the wheels so that they are shifted by the body center amount
    ModelTransform bodyShiftTransform = new ModelTransform(botBody.getAttributeValue("x").getStandardValue(), 
        botBody.getAttributeValue("y").getStandardValue(), botBody.getAttributeValue("z").getStandardValue());
    

    Measurement axisShift = (botLength.divideMD(2.0)).subtract(m_wheelDistanceFromDriveEdge);
    Measurement wheelDistanceFromBody = (botWidth.divideMD(2.0)).add(m_wheelLength.divideMD(2.0)).add(m_wheelGap);
    
    // Compute how many wheels we have:
    // Make sure there are at least 2 wheels:
    int totalWheels = Math.max(1, (int) (Math.floor(botLength.getStandardValue() / wheelRadius.getStandardValue()) - 1));
    Measurement perWheelMass = m_wheelMass.divideMD(totalWheels);
    List<Double> wheelPositions = new ArrayList<Double>();
    
    // Compute the locations of the wheels:
    // one of the wheels is the passive wheels attached to the flipper, so we don't count those and subtract 1 from the wheel count
    Measurement wheelShift = (botLength.subtract(m_wheelDistanceFromDriveEdge).subtract(m_wheelDistanceFromPassiveEdge)).divideMD(totalWheels-1);
    for (int i=0; i<totalWheels-1; i++)
      wheelPositions.add((axisShift.getStandardValue()*-1) + (wheelShift.getStandardValue()* i));

    // add all the wheels:
    for (int i=0; i<wheelPositions.size(); i++) {
      Double wheelPosition = wheelPositions.get(i);
      ModelTransform driveAxis = new ModelTransform(wheelPosition, 0, 0);
      List<ModelObject> wheelObjects = new ArrayList<ModelObject>();

      // left wheel
      ModelPosition leftPosition = new ModelPosition(new Measurement(Units.getMeters(), 0.0), wheelDistanceFromBody, new Measurement(Units.getMeters(), 0.0));
      ModelRotation leftRotation = new ModelQuaternion(0.7071, 0.7071, 0, 0);
      String name = "Wheel"+i;
      wheelObjects.add(new ModelCylinder("left"+name, "", leftPosition, leftRotation, wheelRadius, m_wheelLength, perWheelMass));

      // left joint
      wheelObjects.add(new ModelHingeJoint("botbody", "left"+name, 0d, 0d, 0d, 0d, 1d, 0d, null, null));

      // right wheel
      ModelPosition rightPosition = new ModelPosition(new Measurement(Units.getMeters(), 0.0), wheelDistanceFromBody.multMD(-1), new Measurement(Units.getMeters(), 0.0));
      ModelRotation rightRotation = new ModelQuaternion(0.7071, 0.7071, 0, 0);
      wheelObjects.add(new ModelCylinder("right"+name, "", rightPosition, rightRotation, wheelRadius, m_wheelLength, perWheelMass));

      // right joint
      wheelObjects.add(new ModelHingeJoint("botbody", "right"+name, 0d, 0d, 0d, 0d, 1d, 0d, null, null));

      // add the wheels to the axis and add the axis to the model
      driveAxis.addTransformTargets(wheelObjects);
      bodyShiftTransform.addTransformTarget(driveAxis);
    }
    // add the transform to the model:
    copy.addTopLevelObject(bodyShiftTransform);
    
    
    // add the other wheels:
    // we want to add to the existing flipper axis:
    ModelTransform flipperAxis = (ModelTransform) idMap.get("leftarm").getParentObject();
    // update this transform based on the zShift from above:
    flipperAxis.setZShift(flipperAxis.getZShift()+zDelta.getStandardValue());

    List<ModelObject> wheelObjects2 = new ArrayList<ModelObject>();

    
    // left passive wheel
    ModelPosition leftPosition2 = new ModelPosition(new Measurement(Units.getMeters(), 0.0), wheelDistanceFromBody, new Measurement(Units.getMeters(), 0.0));
    ModelRotation leftRotation2 = new ModelQuaternion(0.7071, 0.7071, 0, 0);
    wheelObjects2.add(new ModelCylinder("leftPassiveWheel", "", leftPosition2, leftRotation2, wheelRadius, m_wheelLength, perWheelMass));
    
    // left passive joint
    wheelObjects2.add(new ModelHingeJoint("botbody", "leftPassiveWheel", 0d, 0d, 0d, 0d, 1d, 0d, null, null));
    
    // right passive wheel
    ModelPosition rightPosition2 = new ModelPosition(new Measurement(Units.getMeters(), 0.0), wheelDistanceFromBody.multMD(-1), new Measurement(Units.getMeters(), 0.0));
    ModelRotation rightRotation2 = new ModelQuaternion(0.7071, 0.7071, 0, 0);
    wheelObjects2.add(new ModelCylinder("rightPassiveWheel", "", rightPosition2, rightRotation2, wheelRadius, m_wheelLength, perWheelMass));
    
    // right passive joint
    wheelObjects2.add(new ModelHingeJoint("botbody", "rightPassiveWheel", 0d, 0d, 0d, 0d, 1d, 0d, null, null));
    
    // add to the transform, and then add this stuff to the model (but not to the top level!)
    flipperAxis.addTransformTargets(wheelObjects2);
    copy.addComponentLevelObjects(wheelObjects2);
    
    
    // now we need to adjust the flipper position:
    ModelBody leftFlipper = idMap.get("leftarm");
    Measurement leftWidth = leftFlipper.getAttributeValue("lengthy");
    ModelBody rightFlipper = idMap.get("rightarm");
    Measurement rightWidth = rightFlipper.getAttributeValue("lengthy");

    // calculate the position the flippers need to be in
    Measurement rightFlipperDistanceFromBody = wheelDistanceFromBody.add((m_wheelLength.divideMD(2.0)).add((rightWidth.divideMD(2.0)).add(m_flipperGap)));
    Measurement leftFlipperDistanceFromBody = wheelDistanceFromBody.add((m_wheelLength.divideMD(2.0)).add((leftWidth.divideMD(2.0)).add(m_flipperGap)));
    // calculate the delta:
    Measurement leftDelta = leftFlipperDistanceFromBody.subtract(leftFlipper.getAttributeValue("y"));
    Measurement rightDelta = rightFlipperDistanceFromBody.multMD(-1.0).subtract(rightFlipper.getAttributeValue("y"));

    leftFlipper.updateRobotAttributeFromDeltas(new RobotAttributeValues("y", leftDelta), copy.getJoints(leftFlipper));
    rightFlipper.updateRobotAttributeFromDeltas(new RobotAttributeValues("y", rightDelta), copy.getJoints(rightFlipper));
    
    
    // return the modified model:
    return copy;
  }
  
  // change the step height
  protected PhysicalModel modifyEnvironment(PhysicalModel copy, Measurement newStepHeight) {
    Map<String, ModelBody> idMap = copy.getIDMap();
    ModelBody step = idMap.get("step");
    Measurement height = step.getAttributeValue("lengthz");
    Measurement heightDelta = newStepHeight.subtract(height);
    step.updateRobotAttributeFromDeltas(new RobotAttributeValues("lengthz", heightDelta), copy.getJoints(step));
    // make sure that the step is actually touching the ground:
    
    Measurement currentZ = step.getAttributeValue("z");
    Measurement deltaZ = (newStepHeight.divideMD(2.0)).subtract(currentZ);
    if (deltaZ.getStandardValue() != 0) {
      step.updateRobotAttributeFromDeltas(new RobotAttributeValues("z", deltaZ), copy.getJoints(step));
    }
    
    // return the modified model:
    return copy;
  }
  
  /**
   * Write the robot model to the specified file in an appropriate format
   * @param file the file to write to
   * @throws Exception if an error writing to the file occurs
   */
  public abstract void writeToFile(File file) throws Exception;
}
