/*
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.view;

import com.bbn.morph.executer.Pair;
import com.bbn.morph.executer.StressInputType;
import com.bbn.morph.executer.StressInputs;

/**Implements a stress function with three linear regions identified by two points. 
 * 
 * Two kinds of stress functions: 
 * 
 * Type A: 
 * -----(first)
 *        \                         
 *         \
 *         (second) --------
 *   
 *         
 * Type B: 
 *         
 *          (second)--------
 *          /
 *         /
 * -------(first)        
 * 
 * Stress value in [0,1] and MaxDouble if in nonviable space
 * Stress Direction is opposite of the slope (+ for A and - for B) because the 
 * direction is supposed to tell us which way we should move in the x axis so that the stress can be reduced.
 * 
 * @author fyaman
 *
 */
public class ThreePieceLinearStressFunction implements StressFunction{
    StressInputType inputLabel;
    Pair first;
    Pair second;
	double slope;
	final static double MAX_STRESS= Double.MAX_VALUE;
	
	
	public ThreePieceLinearStressFunction(StressInputType inputLabel, Pair first,
			Pair second) {
		super();
		this.inputLabel = inputLabel;
		this.first = first;
		this.second = second;
		slope = (second.getY() - first.getY())/(second.getX() - first.getX());
	}




	public StressInputType getInputLabel() {
		return inputLabel;
	}


	@Override
	public Double computeStress(StressInputs inputs) {
		double in= inputs.get(inputLabel);
		if (in<=first.getX())
			 // y= first.getY();
			if(slope>0)
				return 0.0; // no stress
			else
				return MAX_STRESS;
		if (in>=second.getX())
			if(slope>0)
				return MAX_STRESS; // no stress
			else
				return 0.0;   // y= second.getY();
		double stress;
		if (slope>0)
			stress= (in-first.getX())/(second.getX() -first.getX());
		else
			stress= (in-second.getX())/(first.getX() -second.getX());
		
		if (slope>0)
			return -stress;
		else
			return stress;	  
	}




	@Override
	public String description() {
		String desc="Stress Function: Piecewise linear-- flat before and after the two critical points: #1 " + first + " and #2 " + second; 
		return desc;
	}

	
}
